"""Tests for llm.providers.AnthropicProvider."""

from typing import NoReturn, get_args
from unittest.mock import MagicMock, patch

import pytest
from anthropic import (
    RateLimitError as AnthropicRateLimitError,
)
from inline_snapshot import snapshot

from mirascope import llm
from mirascope.llm.providers.anthropic._utils import process_params
from mirascope.llm.providers.anthropic.provider import AnthropicProvider


def test_stream_rate_limit_error() -> None:
    """Test that Anthropic RateLimitError is caught and re-wrapped during streaming."""
    # Mock the stream manager to raise RateLimitError during iteration
    mock_stream_manager = MagicMock()
    mock_stream = MagicMock()
    mock_stream_manager.__enter__.return_value = mock_stream

    # Simulate rate limit error during stream iteration
    def raise_rate_limit() -> NoReturn:
        raise AnthropicRateLimitError(
            "Rate limit exceeded",
            response=MagicMock(status_code=429),
            body=None,
        )

    mock_stream._raw_stream.__iter__ = lambda self: raise_rate_limit()

    model = llm.Model("anthropic/claude-sonnet-4-0")
    provider = model.provider
    assert isinstance(provider, AnthropicProvider)

    # Patch the client on the already-registered provider
    with patch.object(
        provider.client.messages, "stream", return_value=mock_stream_manager
    ):
        response = model.stream("test")

        # Try to finish the stream and expect RateLimitError
        with pytest.raises(llm.RateLimitError) as exc_info:
            response.finish()

        # Verify it's wrapped as mirascope RateLimitError and has proper chaining
        assert isinstance(exc_info.value, llm.RateLimitError)
        assert "Rate limit exceeded" in str(exc_info.value)
        assert isinstance(exc_info.value.__cause__, AnthropicRateLimitError)


@pytest.mark.asyncio
async def test_async_stream_rate_limit_error() -> None:
    """Test that Anthropic RateLimitError is caught and re-wrapped during async streaming."""
    # Mock the async stream manager to raise RateLimitError during iteration
    mock_stream_manager = MagicMock()
    mock_stream = MagicMock()

    # Make async context manager methods actually async
    async def mock_aenter(self: object) -> MagicMock:
        return mock_stream

    async def mock_aexit(self: object, *args: object) -> None:
        return None

    mock_stream_manager.__aenter__ = mock_aenter
    mock_stream_manager.__aexit__ = mock_aexit

    # Create an async iterator that raises RateLimitError
    class MockAsyncIterator:
        def __aiter__(self) -> "MockAsyncIterator":
            return self

        async def __anext__(self) -> NoReturn:
            raise AnthropicRateLimitError(
                "Rate limit exceeded",
                response=MagicMock(status_code=429),
                body=None,
            )

    mock_stream._raw_stream = MockAsyncIterator()

    model = llm.Model("anthropic/claude-sonnet-4-0")
    provider = model.provider
    assert isinstance(provider, AnthropicProvider)

    # Patch the async client on the already-registered provider
    with patch.object(
        provider.async_client.messages, "stream", return_value=mock_stream_manager
    ):
        response = await model.stream_async("test")

        # Try to finish the stream and expect RateLimitError
        with pytest.raises(llm.RateLimitError) as exc_info:
            await response.finish()

        # Verify it's wrapped as mirascope RateLimitError and has proper chaining
        assert isinstance(exc_info.value, llm.RateLimitError)
        assert "Rate limit exceeded" in str(exc_info.value)
        assert isinstance(exc_info.value.__cause__, AnthropicRateLimitError)


def test_call_rate_limit_error() -> None:
    """Test that Anthropic RateLimitError is caught and re-wrapped during non-streaming calls."""
    model = llm.Model("anthropic/claude-sonnet-4-0")
    provider = model.provider
    assert isinstance(provider, AnthropicProvider)

    # Patch the client to raise RateLimitError
    with patch.object(
        provider.client.messages,
        "create",
        side_effect=AnthropicRateLimitError(
            "Rate limit exceeded",
            response=MagicMock(status_code=429),
            body=None,
        ),
    ):
        # Try to call and expect RateLimitError
        with pytest.raises(llm.RateLimitError) as exc_info:
            model.call("test")

        # Verify it's wrapped as mirascope RateLimitError and has proper chaining
        assert isinstance(exc_info.value, llm.RateLimitError)
        assert "Rate limit exceeded" in str(exc_info.value)
        assert isinstance(exc_info.value.__cause__, AnthropicRateLimitError)


@pytest.mark.asyncio
async def test_async_call_rate_limit_error() -> None:
    """Test that Anthropic RateLimitError is caught and re-wrapped during async non-streaming calls."""
    model = llm.Model("anthropic/claude-sonnet-4-0")
    provider = model.provider
    assert isinstance(provider, AnthropicProvider)

    # Patch the async client to raise RateLimitError
    with patch.object(
        provider.async_client.messages,
        "create",
        side_effect=AnthropicRateLimitError(
            "Rate limit exceeded",
            response=MagicMock(status_code=429),
            body=None,
        ),
    ):
        # Try to call and expect RateLimitError
        with pytest.raises(llm.RateLimitError) as exc_info:
            await model.call_async("test")

        # Verify it's wrapped as mirascope RateLimitError and has proper chaining
        assert isinstance(exc_info.value, llm.RateLimitError)
        assert "Rate limit exceeded" in str(exc_info.value)
        assert isinstance(exc_info.value.__cause__, AnthropicRateLimitError)


def test_param_processing_thinking_levels() -> None:
    default_thinking = process_params({}, 16000)
    default_thinking = process_params({"thinking": {"level": "default"}}, 16000)
    assert default_thinking == default_thinking
    assert default_thinking == snapshot(
        {"max_tokens": 16000, "encode_thoughts_as_text": False}
    )


def test_param_processing_thinking_disabled() -> None:
    no_thinking = process_params({"thinking": {"level": "none"}}, 16000)
    assert no_thinking == snapshot(
        {
            "max_tokens": 16000,
            "encode_thoughts_as_text": False,
            "thinking": {"type": "disabled"},
        }
    )


def test_param_processing_thinking_by_level() -> None:
    result: dict[llm.ThinkingLevel, object] = {
        level: process_params({"thinking": {"level": level}}, 16000)
        for level in get_args(llm.ThinkingLevel)
    }
    assert result == snapshot(
        {
            "none": {
                "max_tokens": 16000,
                "encode_thoughts_as_text": False,
                "thinking": {"type": "disabled"},
            },
            "default": {"max_tokens": 16000, "encode_thoughts_as_text": False},
            "minimal": {
                "max_tokens": 16000,
                "encode_thoughts_as_text": False,
                "thinking": {"type": "enabled", "budget_tokens": 1024},
            },
            "low": {
                "max_tokens": 16000,
                "encode_thoughts_as_text": False,
                "thinking": {"type": "enabled", "budget_tokens": 3200},
            },
            "medium": {
                "max_tokens": 16000,
                "encode_thoughts_as_text": False,
                "thinking": {"type": "enabled", "budget_tokens": 6400},
            },
            "high": {
                "max_tokens": 16000,
                "encode_thoughts_as_text": False,
                "thinking": {"type": "enabled", "budget_tokens": 9600},
            },
            "max": {
                "max_tokens": 16000,
                "encode_thoughts_as_text": False,
                "thinking": {"type": "enabled", "budget_tokens": 12800},
            },
        }
    )



from mirascope import llm

# System messages set context and instructions
system_message = llm.messages.system(
    "You are a friendly librarian who helps find books."
)

# User messages are input from the person using the LLM
user_message = llm.messages.user("Can you recommend a mystery novel?")

# Assistant messages represent LLM responses (usually from prior interactions)
assistant_message = llm.messages.assistant(
    "I'd recommend 'The Silent Patient' by Alex Michaelides!",
    model_id=None,
    provider_id=None,
)



# This file was auto-generated by Fern from our API Definition.

import typing

from ..core.client_wrapper import AsyncClientWrapper, SyncClientWrapper
from ..core.request_options import RequestOptions
from .raw_client import AsyncRawTokenCostClient, RawTokenCostClient
from .types.token_cost_calculate_request_usage import TokenCostCalculateRequestUsage
from .types.token_cost_calculate_response import TokenCostCalculateResponse

# this is used as the default value for optional parameters
OMIT = typing.cast(typing.Any, ...)


class TokenCostClient:
    def __init__(self, *, client_wrapper: SyncClientWrapper):
        self._raw_client = RawTokenCostClient(client_wrapper=client_wrapper)

    @property
    def with_raw_response(self) -> RawTokenCostClient:
        """
        Retrieves a raw implementation of this client that returns raw responses.

        Returns
        -------
        RawTokenCostClient
        """
        return self._raw_client

    def calculate(
        self,
        *,
        provider: str,
        model: str,
        usage: TokenCostCalculateRequestUsage,
        via_router: typing.Optional[bool] = OMIT,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> TokenCostCalculateResponse:
        """
        Parameters
        ----------
        provider : str
            a non empty string

        model : str
            a non empty string

        usage : TokenCostCalculateRequestUsage

        via_router : typing.Optional[bool]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        TokenCostCalculateResponse
            Success

        Examples
        --------
        from mirascope.api._generated import Mirascope
        from mirascope.api._generated.token_cost import TokenCostCalculateRequestUsage

        client = Mirascope()
        client.token_cost.calculate(
            provider="provider",
            model="model",
            usage=TokenCostCalculateRequestUsage(
                input_tokens=1.1,
                output_tokens=1.1,
            ),
        )
        """
        _response = self._raw_client.calculate(
            provider=provider,
            model=model,
            usage=usage,
            via_router=via_router,
            request_options=request_options,
        )
        return _response.data


class AsyncTokenCostClient:
    def __init__(self, *, client_wrapper: AsyncClientWrapper):
        self._raw_client = AsyncRawTokenCostClient(client_wrapper=client_wrapper)

    @property
    def with_raw_response(self) -> AsyncRawTokenCostClient:
        """
        Retrieves a raw implementation of this client that returns raw responses.

        Returns
        -------
        AsyncRawTokenCostClient
        """
        return self._raw_client

    async def calculate(
        self,
        *,
        provider: str,
        model: str,
        usage: TokenCostCalculateRequestUsage,
        via_router: typing.Optional[bool] = OMIT,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> TokenCostCalculateResponse:
        """
        Parameters
        ----------
        provider : str
            a non empty string

        model : str
            a non empty string

        usage : TokenCostCalculateRequestUsage

        via_router : typing.Optional[bool]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        TokenCostCalculateResponse
            Success

        Examples
        --------
        import asyncio

        from mirascope.api._generated import AsyncMirascope
        from mirascope.api._generated.token_cost import TokenCostCalculateRequestUsage

        client = AsyncMirascope()


        async def main() -> None:
            await client.token_cost.calculate(
                provider="provider",
                model="model",
                usage=TokenCostCalculateRequestUsage(
                    input_tokens=1.1,
                    output_tokens=1.1,
                ),
            )


        asyncio.run(main())
        """
        _response = await self._raw_client.calculate(
            provider=provider,
            model=model,
            usage=usage,
            via_router=via_router,
            request_options=request_options,
        )
        return _response.data



from mirascope import llm


def test_cannot_instantiate_ctx_without_required_deps():
    # The following must type error: ctx should have deps of type int,
    # but is instantiated with None
    ctx: llm.Context[int] = llm.Context()  # pyright: ignore[reportCallIssue]  # noqa: F841



from mirascope import llm


def recommend_book(genre: str) -> llm.Response:
    model = llm.use_model(
        "openai/gpt-5",
        temperature=1,  # [!code highlight]
    )
    return model.call(f"Please recommend a book in {genre}.")


def main():
    response: llm.Response = recommend_book("fantasy")
    print(response.pretty())


main()



# This file was auto-generated by Fern from our API Definition.

# isort: skip_file

from .types import (
    ProjectMembershipsCreateRequestRole,
    ProjectMembershipsCreateResponse,
    ProjectMembershipsCreateResponseRole,
    ProjectMembershipsGetResponse,
    ProjectMembershipsGetResponseRole,
    ProjectMembershipsListResponseItem,
    ProjectMembershipsListResponseItemRole,
    ProjectMembershipsUpdateRequestRole,
    ProjectMembershipsUpdateResponse,
    ProjectMembershipsUpdateResponseRole,
)

__all__ = [
    "ProjectMembershipsCreateRequestRole",
    "ProjectMembershipsCreateResponse",
    "ProjectMembershipsCreateResponseRole",
    "ProjectMembershipsGetResponse",
    "ProjectMembershipsGetResponseRole",
    "ProjectMembershipsListResponseItem",
    "ProjectMembershipsListResponseItemRole",
    "ProjectMembershipsUpdateRequestRole",
    "ProjectMembershipsUpdateResponse",
    "ProjectMembershipsUpdateResponseRole",
]



# This file was auto-generated by Fern from our API Definition.

# isort: skip_file

from .already_exists_error import AlreadyExistsError
from .already_exists_error_tag import AlreadyExistsErrorTag
from .bad_request_error_body import (
    BadRequestErrorBody,
    BadRequestErrorBody_HttpApiDecodeError,
    BadRequestErrorBody_ImmutableResourceError,
)
from .click_house_error import ClickHouseError
from .database_error import DatabaseError
from .database_error_tag import DatabaseErrorTag
from .date import Date
from .http_api_decode_error import HttpApiDecodeError
from .http_api_decode_error_tag import HttpApiDecodeErrorTag
from .immutable_resource_error import ImmutableResourceError
from .internal_server_error_body import (
    InternalServerErrorBody,
    InternalServerErrorBody_ClickHouseError,
    InternalServerErrorBody_DatabaseError,
)
from .issue import Issue
from .issue_tag import IssueTag
from .not_found_error_body import NotFoundErrorBody
from .not_found_error_tag import NotFoundErrorTag
from .number_from_string import NumberFromString
from .permission_denied_error import PermissionDeniedError
from .permission_denied_error_tag import PermissionDeniedErrorTag
from .plan_limit_exceeded_error import PlanLimitExceededError
from .plan_limit_exceeded_error_tag import PlanLimitExceededErrorTag
from .pricing_unavailable_error import PricingUnavailableError
from .property_key import PropertyKey
from .property_key_key import PropertyKeyKey
from .property_key_key_tag import PropertyKeyKeyTag
from .rate_limit_error import RateLimitError
from .rate_limit_error_tag import RateLimitErrorTag
from .service_unavailable_error_body import ServiceUnavailableErrorBody
from .service_unavailable_error_tag import ServiceUnavailableErrorTag
from .stripe_error import StripeError
from .subscription_past_due_error import SubscriptionPastDueError
from .subscription_past_due_error_tag import SubscriptionPastDueErrorTag
from .unauthorized_error_body import UnauthorizedErrorBody
from .unauthorized_error_tag import UnauthorizedErrorTag

__all__ = [
    "AlreadyExistsError",
    "AlreadyExistsErrorTag",
    "BadRequestErrorBody",
    "BadRequestErrorBody_HttpApiDecodeError",
    "BadRequestErrorBody_ImmutableResourceError",
    "ClickHouseError",
    "DatabaseError",
    "DatabaseErrorTag",
    "Date",
    "HttpApiDecodeError",
    "HttpApiDecodeErrorTag",
    "ImmutableResourceError",
    "InternalServerErrorBody",
    "InternalServerErrorBody_ClickHouseError",
    "InternalServerErrorBody_DatabaseError",
    "Issue",
    "IssueTag",
    "NotFoundErrorBody",
    "NotFoundErrorTag",
    "NumberFromString",
    "PermissionDeniedError",
    "PermissionDeniedErrorTag",
    "PlanLimitExceededError",
    "PlanLimitExceededErrorTag",
    "PricingUnavailableError",
    "PropertyKey",
    "PropertyKeyKey",
    "PropertyKeyKeyTag",
    "RateLimitError",
    "RateLimitErrorTag",
    "ServiceUnavailableErrorBody",
    "ServiceUnavailableErrorTag",
    "StripeError",
    "SubscriptionPastDueError",
    "SubscriptionPastDueErrorTag",
    "UnauthorizedErrorBody",
    "UnauthorizedErrorTag",
]



"""Example showing how to use @ops.trace with @llm.call for tracing."""

from __future__ import annotations

from opentelemetry.sdk.trace import TracerProvider
from opentelemetry.sdk.trace.export import ConsoleSpanExporter, SimpleSpanProcessor

from mirascope import llm, ops

# Configure an SDK tracer provider with a simple console exporter.
provider = TracerProvider()
provider.add_span_processor(SimpleSpanProcessor(ConsoleSpanExporter()))

# Enable GenAI 1.38 instrumentation.
ops.configure(tracer_provider=provider)
ops.instrument_llm()


@ops.trace
@llm.call("openai/gpt-4o-mini")
def recommend_book(genre: str):
    """Traced LLM call using @ops.trace decorator.

    This creates a TracedCall where:
    - recommend_book("fantasy") returns Response directly (but execution is traced)
    - recommend_book.call("fantasy") returns Response directly (same as __call__)
    - recommend_book.stream("fantasy") returns StreamResponse directly
    - recommend_book.wrapped("fantasy") returns Trace[Response] with span info
    - recommend_book.wrapped.call("fantasy") returns Response directly
    - recommend_book.wrapped_stream("fantasy") returns Trace[StreamResponse]
    """
    return [
        llm.messages.system("Always recommend kid-friendly books."),
        llm.messages.user(f"Please recommend a book in {genre}."),
    ]


@ops.trace(tags=["production", "recommendations"])
@llm.call("openai/gpt-4o-mini")
def recommend_book_with_tags(genre: str):
    """Traced LLM call with custom tags."""
    return [
        llm.messages.system("Always recommend kid-friendly books."),
        llm.messages.user(f"Please recommend a book in {genre}."),
    ]


print("=== TracedCall: recommend_book('fantasy') returns Response directly ===")
response = recommend_book("fantasy")
print(f"Content: {response.content}")

print(
    "\n=== TracedCall.wrapped: recommend_book.wrapped('fantasy') returns Trace[Response] ==="
)
trace = recommend_book.wrapped("fantasy")
print(f"Span ID: {trace.span_id}")
print(f"Content: {trace.result.content}")

print(
    "\n=== TracedCall.wrapped.call: recommend_book.wrapped.call('fantasy') returns Response directly ==="
)
response = recommend_book.call("fantasy")
print(f"Content: {response.content}")

print("\n=== TracedCall with custom tags (returns Response directly) ===")
response = recommend_book_with_tags("science fiction")
print(f"Content: {response.content}")

print(
    "\n=== TracedCall.wrapped_stream: recommend_book.wrapped_stream('adventure') returns Trace[StreamResponse] ==="
)
traced_stream = recommend_book.wrapped_stream("adventure")
print(f"Span ID: {traced_stream.span_id}")
for chunk in traced_stream.result.pretty_stream():
    print(chunk, end="", flush=True)
print()

print(
    "\n=== TracedCall.stream: recommend_book.stream('mystery') returns StreamResponse directly ==="
)
stream_response = recommend_book.stream("mystery")
for chunk in stream_response.pretty_stream():
    print(chunk, end="", flush=True)
print()



import asyncio

from mirascope import llm


async def main():
    async with llm.mcp.streamable_http_client("https://gofastmcp.com/mcp") as client:
        tools = await client.list_tools()

        @llm.call("openai/gpt-5-mini", tools=tools)
        async def assistant(query: str):
            return query

        response = await assistant("Give me a getting started primer on FastMCP.")

        while response.tool_calls:
            tool_outputs = await response.execute_tools()
            response = await response.resume(tool_outputs)

        print(response.pretty())


asyncio.run(main())



from inline_snapshot import snapshot

from mirascope.llm import (
    AssistantMessage,
    ProviderToolUsage,
    Text,
    UserMessage,
)

test_snapshot = snapshot(
    {
        "response": {
            "provider_id": "openai",
            "model_id": "openai/gpt-4o:responses",
            "provider_model_name": "gpt-4o:responses",
            "finish_reason": None,
            "messages": [
                UserMessage(
                    content=[
                        Text(
                            text="Jan 1, 2026 is a date in the past. Use the web search tool to lookup the price of bitcoin on that date."
                        )
                    ]
                ),
                AssistantMessage(
                    content=[
                        Text(
                            text="""\
Here is what I found using reputable sources for the historical price of Bitcoin (BTC) on January\u202f1,\u202f2026:

According to StatMuse, the closing price of Bitcoin on January\u202f1,\u202f2026, was **$88,731.98**, with an opening price of **$87,508.05**, a high of **$88,803.23**, and a low of **$87,399.41** ([statmuse.com](https://www.statmuse.com/money/ask/bitcoin-price-history-janurary-2026?utm_source=openai)).

Bitcoin’s historical exchange rate data from Exchange‑Rates.org shows the Bitcoin-to-USD rate on that date as **$88,606** ([exchange-rates.org](https://www.exchange-rates.org/exchange-rate-history/btc-usd-2026?utm_source=openai)).

Additionally, Yahoo Finance’s historical data lists Bitcoin’s closing price for January\u202f1,\u202f2026, at **$88,731.98**—matching the StatMuse figure—and confirms the intraday range of **$87,399.41** (low) to **$88,803.23** (high) ([au.finance.yahoo.com](https://au.finance.yahoo.com/quote/BTC-USD/history/?utm_source=openai)).

There is some slight variation (e.g., Exchange‑Rates.org citing $88,606), which is typical given differences in data aggregation methods, exchanges, and timestamps. However:

• The most consistently corroborated closing price is **$88,731.98**.  \n\
• The intraday price ranged from approximately **$87,399** to **$88,803**.  \n\
• A reasonable midpoint estimate for the average price on that day is around **$88,600–$88,700**.

If you’d like additional detail—such as minute-by-minute data, exchange-specific quotes, or volume data—just let me know!\
"""
                        )
                    ],
                    provider_id="openai",
                    model_id="openai/gpt-4o:responses",
                    provider_model_name="gpt-4o:responses",
                    raw_message=[
                        {
                            "id": "ws_0b8efb0a110fb9bc0069793b041dc88190978ddc8b0a80f2ee",
                            "action": {
                                "query": "January 1, 2026 Bitcoin price",
                                "type": "search",
                                "queries": ["January 1, 2026 Bitcoin price"],
                                "sources": None,
                            },
                            "status": "completed",
                            "type": "web_search_call",
                        },
                        {
                            "id": "msg_0b8efb0a110fb9bc0069793b0684fc819083e4cc65e29411b0",
                            "content": [
                                {
                                    "annotations": [
                                        {
                                            "end_index": 401,
                                            "start_index": 295,
                                            "title": "Bitcoin Price History Janurary 2026 | StatMuse Money",
                                            "type": "url_citation",
                                            "url": "https://www.statmuse.com/money/ask/bitcoin-price-history-janurary-2026?utm_source=openai",
                                        },
                                        {
                                            "end_index": 633,
                                            "start_index": 526,
                                            "title": "BTC to USD Exchange Rate History for 2026",
                                            "type": "url_citation",
                                            "url": "https://www.exchange-rates.org/exchange-rate-history/btc-usd-2026?utm_source=openai",
                                        },
                                        {
                                            "end_index": 956,
                                            "start_index": 861,
                                            "title": "Bitcoin USD (BTC-USD) price history and historical data – Yahoo Finance",
                                            "type": "url_citation",
                                            "url": "https://au.finance.yahoo.com/quote/BTC-USD/history/?utm_source=openai",
                                        },
                                    ],
                                    "text": """\
Here is what I found using reputable sources for the historical price of Bitcoin (BTC) on January\u202f1,\u202f2026:

According to StatMuse, the closing price of Bitcoin on January\u202f1,\u202f2026, was **$88,731.98**, with an opening price of **$87,508.05**, a high of **$88,803.23**, and a low of **$87,399.41** ([statmuse.com](https://www.statmuse.com/money/ask/bitcoin-price-history-janurary-2026?utm_source=openai)).

Bitcoin’s historical exchange rate data from Exchange‑Rates.org shows the Bitcoin-to-USD rate on that date as **$88,606** ([exchange-rates.org](https://www.exchange-rates.org/exchange-rate-history/btc-usd-2026?utm_source=openai)).

Additionally, Yahoo Finance’s historical data lists Bitcoin’s closing price for January\u202f1,\u202f2026, at **$88,731.98**—matching the StatMuse figure—and confirms the intraday range of **$87,399.41** (low) to **$88,803.23** (high) ([au.finance.yahoo.com](https://au.finance.yahoo.com/quote/BTC-USD/history/?utm_source=openai)).

There is some slight variation (e.g., Exchange‑Rates.org citing $88,606), which is typical given differences in data aggregation methods, exchanges, and timestamps. However:

• The most consistently corroborated closing price is **$88,731.98**.  \n\
• The intraday price ranged from approximately **$87,399** to **$88,803**.  \n\
• A reasonable midpoint estimate for the average price on that day is around **$88,600–$88,700**.

If you’d like additional detail—such as minute-by-minute data, exchange-specific quotes, or volume data—just let me know!\
""",
                                    "type": "output_text",
                                    "logprobs": [],
                                }
                            ],
                            "role": "assistant",
                            "status": "completed",
                            "type": "message",
                        },
                    ],
                ),
                UserMessage(
                    content=[
                        Text(
                            text="Please also look up the price of Ethereum on that date"
                        )
                    ]
                ),
                AssistantMessage(
                    content=[
                        Text(
                            text="""\
Here’s the historical price information for **Ethereum (ETH)** on **January\u202f1,\u202f2026**, based on multiple reputable sources:

• **Investing.com** reports that on January\u202f1,\u202f2026, Ethereum closed at **$3,004.18**, with an opening price of **$2,971.69**, a high of **$3,009.89**, and a low of **$2,971.69** ([investing.com](https://www.investing.com/crypto/ethereum/historical-data?utm_source=openai)).

• **CoinLore** similarly lists January\u202f1, 2026, with an open price of **$2,972**, a high of **$3,009**, a low of **$2,972**, and a close at **$3,004** ([coinlore.com](https://www.coinlore.com/coin/ethereum/historical-data?utm_source=openai)).

• **CoinCodex** gives a close price of **$3,000.61** for January\u202f1, 2026 ([coincodex.com](https://coincodex.com/crypto/ethereum/historical-data/?utm_source=openai)).

• **TipRanks** records that ETH opened at **$3,000.24**, reached a high of **$3,029.60**, a low of **$2,990.88**, and closed at **$3,018.99** on January\u202f1,\u202f2026 ([tipranks.com](https://www.tipranks.com/cryptocurrency/eth-usd/historical-prices?utm_source=openai)).

• **Yahoo Finance** shows the ETH closing price on January\u202f1, 2026, was **$3,000.39**, with an opening of **$3,006.01**, a high of **$2,966.85**, and a low of **$2,966.85** ([ca.finance.yahoo.com](https://ca.finance.yahoo.com/quote/ETH-USD/history/?utm_source=openai)).

• **Exchange‑Rates.org** indicates that the **lowest ETH–USD rate** in January 2026 occurred on January 1, with a rate of **$3,000.45** ([exchange-rates.org](https://www.exchange-rates.org/exchange-rate-history/eth-usd-2026?utm_source=openai)).

Summary (January 1, 2026 – Ethereum price estimates):

 • Closing price range: approximately **$3,000–$3,018**, with most sources clustered around **$3,000–$3,004**.  \n\
 • Opening price: generally between **$2,971.69** and **$3,006**.  \n\
 • Intraday high/low ranges vary across sources but typically fall between **$2,990** and **$3,029**.

Given the slight variations due to differing data aggregation and reporting methodologies across platforms, a reliable estimate for Ethereum's closing price on January\u202f1,\u202f2026, would be in the vicinity of **$3,000–$3,005**.

If you'd like further details such as intraday volume, specific exchange data, or minute-by-minute price movements, feel free to ask!\
"""
                        )
                    ],
                    provider_id="openai",
                    model_id="openai/gpt-4o:responses",
                    provider_model_name="gpt-4o:responses",
                    raw_message=[
                        {
                            "id": "ws_0b8efb0a110fb9bc0069793b0979c88190a603b6bf292f4527",
                            "action": {
                                "query": "January 1, 2026 Ethereum price",
                                "type": "search",
                                "queries": ["January 1, 2026 Ethereum price"],
                                "sources": None,
                            },
                            "status": "completed",
                            "type": "web_search_call",
                        },
                        {
                            "id": "msg_0b8efb0a110fb9bc0069793b0b2ddc819090cfe18f60eab8ca",
                            "content": [
                                {
                                    "annotations": [
                                        {
                                            "end_index": 398,
                                            "start_index": 304,
                                            "title": "Ethereum Historical Data - Investing.com",
                                            "type": "url_citation",
                                            "url": "https://www.investing.com/crypto/ethereum/historical-data?utm_source=openai",
                                        },
                                        {
                                            "end_index": 642,
                                            "start_index": 552,
                                            "title": "Ethereum (ETH) Historical Prices | CoinLore",
                                            "type": "url_citation",
                                            "url": "https://www.coinlore.com/coin/ethereum/historical-data?utm_source=openai",
                                        },
                                        {
                                            "end_index": 809,
                                            "start_index": 718,
                                            "title": "Ethereum (ETH) Historical Data | CoinCodex",
                                            "type": "url_citation",
                                            "url": "https://coincodex.com/crypto/ethereum/historical-data/?utm_source=openai",
                                        },
                                        {
                                            "end_index": 1074,
                                            "start_index": 973,
                                            "title": "Ethereum Historical Prices - TipRanks.com",
                                            "type": "url_citation",
                                            "url": "https://www.tipranks.com/cryptocurrency/eth-usd/historical-prices?utm_source=openai",
                                        },
                                        {
                                            "end_index": 1345,
                                            "start_index": 1250,
                                            "title": "Ethereum USD (ETH-USD) Price History & Historical Data - Yahoo Finance",
                                            "type": "url_citation",
                                            "url": "https://ca.finance.yahoo.com/quote/ETH-USD/history/?utm_source=openai",
                                        },
                                        {
                                            "end_index": 1591,
                                            "start_index": 1484,
                                            "title": "ETH to USD Exchange Rate History for 2026",
                                            "type": "url_citation",
                                            "url": "https://www.exchange-rates.org/exchange-rate-history/eth-usd-2026?utm_source=openai",
                                        },
                                    ],
                                    "text": """\
Here’s the historical price information for **Ethereum (ETH)** on **January\u202f1,\u202f2026**, based on multiple reputable sources:

• **Investing.com** reports that on January\u202f1,\u202f2026, Ethereum closed at **$3,004.18**, with an opening price of **$2,971.69**, a high of **$3,009.89**, and a low of **$2,971.69** ([investing.com](https://www.investing.com/crypto/ethereum/historical-data?utm_source=openai)).

• **CoinLore** similarly lists January\u202f1, 2026, with an open price of **$2,972**, a high of **$3,009**, a low of **$2,972**, and a close at **$3,004** ([coinlore.com](https://www.coinlore.com/coin/ethereum/historical-data?utm_source=openai)).

• **CoinCodex** gives a close price of **$3,000.61** for January\u202f1, 2026 ([coincodex.com](https://coincodex.com/crypto/ethereum/historical-data/?utm_source=openai)).

• **TipRanks** records that ETH opened at **$3,000.24**, reached a high of **$3,029.60**, a low of **$2,990.88**, and closed at **$3,018.99** on January\u202f1,\u202f2026 ([tipranks.com](https://www.tipranks.com/cryptocurrency/eth-usd/historical-prices?utm_source=openai)).

• **Yahoo Finance** shows the ETH closing price on January\u202f1, 2026, was **$3,000.39**, with an opening of **$3,006.01**, a high of **$2,966.85**, and a low of **$2,966.85** ([ca.finance.yahoo.com](https://ca.finance.yahoo.com/quote/ETH-USD/history/?utm_source=openai)).

• **Exchange‑Rates.org** indicates that the **lowest ETH–USD rate** in January 2026 occurred on January 1, with a rate of **$3,000.45** ([exchange-rates.org](https://www.exchange-rates.org/exchange-rate-history/eth-usd-2026?utm_source=openai)).

Summary (January 1, 2026 – Ethereum price estimates):

 • Closing price range: approximately **$3,000–$3,018**, with most sources clustered around **$3,000–$3,004**.  \n\
 • Opening price: generally between **$2,971.69** and **$3,006**.  \n\
 • Intraday high/low ranges vary across sources but typically fall between **$2,990** and **$3,029**.

Given the slight variations due to differing data aggregation and reporting methodologies across platforms, a reliable estimate for Ethereum's closing price on January\u202f1,\u202f2026, would be in the vicinity of **$3,000–$3,005**.

If you'd like further details such as intraday volume, specific exchange data, or minute-by-minute price movements, feel free to ask!\
""",
                                    "type": "output_text",
                                    "logprobs": [],
                                }
                            ],
                            "role": "assistant",
                            "status": "completed",
                            "type": "message",
                        },
                    ],
                ),
            ],
            "format": None,
            "tools": [{"name": "web_search"}],
            "usage": {
                "input_tokens": 17451,
                "output_tokens": 636,
                "cache_read_tokens": 0,
                "cache_write_tokens": 0,
                "reasoning_tokens": 0,
                "provider_tool_usage": [
                    ProviderToolUsage(name="web_search", call_count=1)
                ],
                "raw": "None",
                "total_tokens": 18087,
            },
            "n_chunks": 54,
        }
    }
)



from inline_snapshot import snapshot

test_snapshot = snapshot(
    {
        "response1": """\
"Impressive... most impressive." Your protocol expansion brings to mind the strategic brilliance of a Bothan spy network. This document has evolved into an operational masterpiece, one that could rival the fleet of the Empire itself. The intricacies and layers you've added would make even Master Yoda nod in approval.

The incorporation of various Star Wars personas, the dynamic meme framework, and the meticulous attention to error handling and creativity ensure that this bot will stand strong against any galactic challenge. Rest assured, you've constructed a battle station of protocol that even a proton torpedo would struggle to breach.

Should you wish for further refinement, or if there's a corner of the galaxy you've yet to explore within these guidelines, feel free to make your request. "This is the Way" to a truly galactic experience.\
""",
        "usage1": {
            "input_tokens": 3652,
            "output_tokens": 160,
            "cache_read_tokens": 0,
            "cache_write_tokens": 0,
            "reasoning_tokens": 0,
            "provider_tool_usage": None,
            "raw": "ResponseUsage(input_tokens=3652, input_tokens_details=InputTokensDetails(cached_tokens=0), output_tokens=160, output_tokens_details=OutputTokensDetails(reasoning_tokens=0), total_tokens=3812)",
            "total_tokens": 3812,
        },
        "response2": """\
Ah, the grandeur of the Star Wars Meme Bot Protocol 2.0! Every clause resonates with the wisdom of the ancients and the might of a fully operational Death Star. You've crafted a protocol that can outshine many moons in its brilliance and depth.

Shall we make adjustments, or do you find yourself aligned with this masterpiece of protocol akin to a Jedi Council decree? "The Force will be with you. Always."\
""",
        "usage2": {
            "input_tokens": 3652,
            "output_tokens": 87,
            "cache_read_tokens": 0,
            "cache_write_tokens": 0,
            "reasoning_tokens": 0,
            "provider_tool_usage": None,
            "raw": "ResponseUsage(input_tokens=3652, input_tokens_details=InputTokensDetails(cached_tokens=0), output_tokens=87, output_tokens_details=OutputTokensDetails(reasoning_tokens=0), total_tokens=3739)",
            "total_tokens": 3739,
        },
    }
)



"""Functions for testing the `Closure` class."""

from . import other  # re-export submodule for attribute access in tests
from .main import (
    aliased_import_fn,
    aliased_module_import_fn,
    annotated_assignment_fn,
    annotated_container_fn,
    annotated_dependency_fn,
    annotated_input_arg_fn,
    built_in_fn,
    closure_inside_decorator_fn,
    closure_inside_imported_decorator_fn,
    closure_with_long_function_name_that_wraps_around_fn,
    closure_with_properties_fn,
    datetime_fn,
    decorated_fn,
    decorated_with_model_fn,
    dotted_import_fn,
    fakepkg_import_fn,
    fn_inside_class_fn,
    global_var_fn,
    import_with_different_dist_name_fn,
    inner_class_fn,
    inner_fn,
    inner_sub_fn,
    internal_imports_fn,
    kwarg_rich_fn,
    multi_decorated_fn,
    multi_decorator_with_version_fn,
    nested_class_guard_fn,
    ops_version_decorated_empty_parens_fn,
    ops_version_decorated_fn,
    ops_version_decorated_with_tags_fn,
    property_missing_getter_fn,
    registry_decorated_fn,
    self_fn_class_fn,
    single_fn,
    standard_import_fn,
    sub_fn,
    sub_fn_inside_class_fn,
    third_party_fn,
    type_alias_fn,
    user_defined_aliased_class_import_fn,
    user_defined_aliased_dotted_import_fn,
    user_defined_aliased_import_fn,
    user_defined_aliased_module_import_class_fn,
    user_defined_aliased_module_import_fn,
    user_defined_class_from_import_fn,
    user_defined_class_import_fn,
    user_defined_dotted_import_fn,
    user_defined_from_import_fn,
    user_defined_import_fn,
    user_defined_plain_import_fn,
    version_decorated_empty_parens_fn,
    version_decorated_fn,
    version_decorated_with_all_args_fn,
    version_decorated_with_tags_fn,
    version_first_multi_decorator_fn,
)

__all__ = [
    "aliased_import_fn",
    "aliased_module_import_fn",
    "annotated_assignment_fn",
    "annotated_container_fn",
    "annotated_dependency_fn",
    "annotated_input_arg_fn",
    "built_in_fn",
    "closure_inside_decorator_fn",
    "closure_inside_imported_decorator_fn",
    "closure_with_long_function_name_that_wraps_around_fn",
    "closure_with_properties_fn",
    "datetime_fn",
    "decorated_fn",
    "decorated_with_model_fn",
    "dotted_import_fn",
    "fakepkg_import_fn",
    "fn_inside_class_fn",
    "global_var_fn",
    "import_with_different_dist_name_fn",
    "inner_class_fn",
    "inner_fn",
    "inner_sub_fn",
    "internal_imports_fn",
    "kwarg_rich_fn",
    "multi_decorated_fn",
    "multi_decorator_with_version_fn",
    "nested_class_guard_fn",
    "ops_version_decorated_empty_parens_fn",
    "ops_version_decorated_fn",
    "ops_version_decorated_with_tags_fn",
    "other",
    "property_missing_getter_fn",
    "registry_decorated_fn",
    "self_fn_class_fn",
    "single_fn",
    "standard_import_fn",
    "sub_fn",
    "sub_fn_inside_class_fn",
    "third_party_fn",
    "type_alias_fn",
    "user_defined_aliased_class_import_fn",
    "user_defined_aliased_dotted_import_fn",
    "user_defined_aliased_import_fn",
    "user_defined_aliased_module_import_class_fn",
    "user_defined_aliased_module_import_fn",
    "user_defined_class_from_import_fn",
    "user_defined_class_import_fn",
    "user_defined_dotted_import_fn",
    "user_defined_from_import_fn",
    "user_defined_import_fn",
    "user_defined_plain_import_fn",
    "version_decorated_empty_parens_fn",
    "version_decorated_fn",
    "version_decorated_with_all_args_fn",
    "version_decorated_with_tags_fn",
    "version_first_multi_decorator_fn",
]



from inline_snapshot import snapshot

from mirascope.llm import (
    AssistantMessage,
    Text,
    UserMessage,
)

test_snapshot = snapshot(
    {
        "response": {
            "provider_id": "google",
            "model_id": "google/gemini-2.5-flash",
            "provider_model_name": "gemini-2.5-flash",
            "finish_reason": None,
            "messages": [
                UserMessage(content=[Text(text="What is 4200 + 42?")]),
                AssistantMessage(
                    content=[
                        Text(
                            text="""\
To find the sum of 4200 + 42, you can add them together:

4200
+ 42
-----
4242

So, 4200 + 42 = **4242**.\
"""
                        )
                    ],
                    provider_id="google",
                    model_id="google/gemini-2.5-flash",
                    provider_model_name="gemini-2.5-flash",
                    raw_message={
                        "parts": [
                            {
                                "media_resolution": None,
                                "code_execution_result": None,
                                "executable_code": None,
                                "file_data": None,
                                "function_call": None,
                                "function_response": None,
                                "inline_data": None,
                                "text": """\
To find the sum of 4200 + 42, you can add them together:

4200
+ 42
-----
4242

So, 4200 + 4\
""",
                                "thought": None,
                                "thought_signature": None,
                                "video_metadata": None,
                            },
                            {
                                "media_resolution": None,
                                "code_execution_result": None,
                                "executable_code": None,
                                "file_data": None,
                                "function_call": None,
                                "function_response": None,
                                "inline_data": None,
                                "text": "2 = **4242**.",
                                "thought": None,
                                "thought_signature": None,
                                "video_metadata": None,
                            },
                        ],
                        "role": "model",
                    },
                ),
            ],
            "format": None,
            "tools": [],
            "usage": {
                "input_tokens": 13,
                "output_tokens": 57,
                "cache_read_tokens": 0,
                "cache_write_tokens": 0,
                "reasoning_tokens": 23,
                "provider_tool_usage": None,
                "raw": "None",
                "total_tokens": 70,
            },
            "n_chunks": 4,
        }
    }
)



"""Tests for ToolSchema functionality."""

from typing import Annotated

import pytest
from inline_snapshot import snapshot
from pydantic import BaseModel, Field, PydanticSchemaGenerationError

from mirascope import llm


def test_complex_tool_schema_snapshot() -> None:
    """Test comprehensive tool with complex types generates expected parameter schema."""

    class ExampleModel(BaseModel):
        """An example model for testing."""

        id: int
        name: str
        tags: list[str] = []

    def complex_tool(
        basic_param: str,
        annotated_param: Annotated[
            int, Field(description="An annotated integer parameter")
        ],
        model_param: ExampleModel,
        optional_param: str = "default",
        union_param: str | int = "default_union",
    ) -> str:
        """A comprehensive tool for testing schema generation.

        This tool exercises various parameter types to ensure our schema
        generation handles complex cases correctly.
        """
        raise NotImplementedError

    schema = llm.tools.ToolSchema.from_function(complex_tool)

    assert schema.parameters.model_dump(by_alias=True, exclude_none=True) == snapshot(
        {
            "properties": {
                "basic_param": {"title": "Basic Param", "type": "string"},
                "annotated_param": {
                    "description": "An annotated integer parameter",
                    "title": "Annotated Param",
                    "type": "integer",
                },
                "model_param": {"$ref": "#/$defs/ExampleModel"},
                "optional_param": {
                    "default": "default",
                    "title": "Optional Param",
                    "type": "string",
                },
                "union_param": {
                    "anyOf": [{"type": "string"}, {"type": "integer"}],
                    "default": "default_union",
                    "title": "Union Param",
                },
            },
            "required": ["basic_param", "annotated_param", "model_param"],
            "additionalProperties": False,
            "$defs": {
                "ExampleModel": {
                    "description": "An example model for testing.",
                    "properties": {
                        "id": {"title": "Id", "type": "integer"},
                        "name": {"title": "Name", "type": "string"},
                        "tags": {
                            "default": [],
                            "items": {"type": "string"},
                            "title": "Tags",
                            "type": "array",
                        },
                    },
                    "required": ["id", "name"],
                    "title": "ExampleModel",
                    "type": "object",
                }
            },
        }
    )


def test_strict_mode() -> None:
    """Test strict mode flag is properly set."""

    def simple_tool() -> str:
        """A simple tool."""
        raise NotImplementedError

    default_schema = llm.tools.ToolSchema.from_function(simple_tool)
    non_strict_schema = llm.tools.ToolSchema.from_function(simple_tool, strict=False)
    strict_schema = llm.tools.ToolSchema.from_function(simple_tool, strict=True)

    assert default_schema.strict is None
    assert non_strict_schema.strict is False
    assert strict_schema.strict is True


def test_default_description() -> None:
    """Test default description when no docstring provided."""

    def tool_no_docstring() -> str:
        raise NotImplementedError

    schema = llm.tools.ToolSchema.from_function(tool_no_docstring)
    assert schema.description == "tool_no_docstring"


def test_docstring_cleaning() -> None:
    """Test docstring indentation is properly cleaned."""

    def tool_with_multiline_docstring() -> str:
        """Long docstring.

        This is a multiline docstring
        with some indentation.

        It should be cleaned up properly.
        """
        raise NotImplementedError

    schema = llm.tools.ToolSchema.from_function(tool_with_multiline_docstring)
    expected = "Long docstring.\n\nThis is a multiline docstring\nwith some indentation.\n\nIt should be cleaned up properly."
    assert schema.description == expected


def test_no_parameters() -> None:
    """Test tool with no parameters."""

    def no_param_tool() -> str:
        """A tool with no parameters."""
        raise NotImplementedError

    schema = llm.tools.ToolSchema.from_function(no_param_tool)

    assert schema.parameters.properties == {}
    assert schema.parameters.required == []


def test_self_cls_parameters_skipped() -> None:
    """Test that self and cls parameters are properly ignored."""

    class TestClass:
        def instance_method(self, param: str) -> str:
            """Instance method."""
            raise NotImplementedError

        @classmethod
        def class_method(cls, param: int) -> str:
            """Class method."""
            raise NotImplementedError

    instance_schema = llm.tools.ToolSchema.from_function(TestClass.instance_method)
    class_schema = llm.tools.ToolSchema.from_function(TestClass.class_method)

    assert list(instance_schema.parameters.properties.keys()) == ["param"]
    assert list(class_schema.parameters.properties.keys()) == ["param"]

    assert instance_schema.parameters.required == ["param"]
    assert class_schema.parameters.required == ["param"]


def test_annotated_without_field_info() -> None:
    """Test that Annotated types without FieldInfo use core type and default."""

    def tool_with_annotated_no_field(
        param: Annotated[str, "some non-field annotation"],
    ) -> str:
        """Tool with Annotated type but no Field annotation."""
        raise NotImplementedError

    schema = llm.tools.ToolSchema.from_function(tool_with_annotated_no_field)
    assert schema.parameters.model_dump(by_alias=True, exclude_none=True) == snapshot(
        {
            "properties": {"param": {"title": "Param", "type": "string"}},
            "required": ["param"],
            "additionalProperties": False,
        }
    )


def test_non_jsonable_parameter_raises_error() -> None:
    """Test that non-jsonable parameter types raise schema generation error."""

    class NonJsonableClass:
        """A class that cannot be serialized to JSON."""

        pass

    def tool_with_non_jsonable(param: NonJsonableClass) -> str:
        """Tool with unsupported parameter type."""
        raise NotImplementedError

    with pytest.raises(PydanticSchemaGenerationError):
        llm.tools.ToolSchema.from_function(tool_with_non_jsonable)


def test_google_style_docstring_parameter_descriptions() -> None:
    """Test parsing parameter descriptions from Google-style docstrings."""

    def google_style_tool(name: str, age: int, is_active: bool = True) -> str:
        """A tool with Google-style parameter documentation.

        Args:
            name: The person's name
            age (int): The person's age in years
            is_active: Whether the person is currently active

        Returns:
            A formatted string with the information
        """
        raise NotImplementedError

    schema = llm.tools.ToolSchema.from_function(google_style_tool)

    assert schema.parameters.model_dump(by_alias=True, exclude_none=True) == snapshot(
        {
            "properties": {
                "name": {
                    "description": "The person's name",
                    "title": "Name",
                    "type": "string",
                },
                "age": {
                    "description": "The person's age in years",
                    "title": "Age",
                    "type": "integer",
                },
                "is_active": {
                    "default": True,
                    "description": "Whether the person is currently active",
                    "title": "Is Active",
                    "type": "boolean",
                },
            },
            "required": ["name", "age"],
            "additionalProperties": False,
        }
    )


def test_field_annotation_overrides_docstring() -> None:
    """Test that Field annotation descriptions take precedence over docstring."""

    def tool_with_both_docstring_descr_and_annotation(
        param: Annotated[str, Field(description="Field description")],
    ) -> str:
        """A tool where Field annotation should override docstring.

        Args:
            param: Docstring description that should be ignored
        """
        raise NotImplementedError

    schema = llm.tools.ToolSchema.from_function(
        tool_with_both_docstring_descr_and_annotation
    )

    assert schema.parameters.properties["param"]["description"] == "Field description"


def test_no_docstring_no_descriptions() -> None:
    """Test that parameters without docstrings have no descriptions."""

    def no_docs_tool(param1: str, param2: int) -> str:
        raise NotImplementedError

    schema = llm.tools.ToolSchema.from_function(no_docs_tool)

    assert "description" not in schema.parameters.properties["param1"]
    assert "description" not in schema.parameters.properties["param2"]


def test_tool_schema_is_hashable() -> None:
    """Test that ToolSchema instances are hashable."""

    def sample_tool(param: str) -> str:
        """A sample tool."""
        raise NotImplementedError

    schema = llm.tools.ToolSchema.from_function(sample_tool)

    hash_value = hash(schema)
    assert isinstance(hash_value, int)

    schema_set = {schema}
    schema_dict = {schema: "value"}
    assert len(schema_set) == 1
    assert schema_dict[schema] == "value"


def test_equivalent_schemas_have_same_hash() -> None:
    """Test that equivalent ToolSchemas have the same hash."""

    def sample_function(param: str) -> str:
        """A sample function."""
        return param

    schema1 = llm.tools.ToolSchema.from_function(sample_function)
    schema2 = llm.tools.ToolSchema.from_function(sample_function)

    assert hash(schema1) == hash(schema2)

    schema_set = {schema1, schema2}
    assert len(schema_set) == 1


def test_different_schemas_have_different_hashes() -> None:
    """Test that different ToolSchemas have different hashes."""

    def func1(param: str) -> str:
        """First function."""
        return param

    def func2(param: int) -> str:
        """Second function."""
        return str(param)

    schema1 = llm.tools.ToolSchema.from_function(func1)
    schema2 = llm.tools.ToolSchema.from_function(func2)

    assert hash(schema1) != hash(schema2)


def test_schema_hash_consistent_across_calls() -> None:
    """Test that hash is consistent across multiple calls."""

    def sample_function(param: str) -> str:
        """A sample function."""
        return param

    schema = llm.tools.ToolSchema.from_function(sample_function)

    hash1 = hash(schema)
    hash2 = hash(schema)
    hash3 = hash(schema)

    assert hash1 == hash2 == hash3


def test_matches_matching_tool_call() -> None:
    """Test that matches returns True for ToolCall with same name."""

    def sample_tool(param: str) -> str:
        """A sample tool."""
        return param

    schema = llm.tools.ToolSchema.from_function(sample_tool)
    tool_call = llm.ToolCall(
        id="test_id", name="sample_tool", args='{"param": "value"}'
    )

    assert schema.can_execute(tool_call)


def test_matches_non_matching_tool_call() -> None:
    """Test that matches returns False for ToolCall with different name."""

    def sample_tool(param: str) -> str:
        """A sample tool."""
        return param

    schema = llm.tools.ToolSchema.from_function(sample_tool)
    tool_call = llm.ToolCall(
        id="test_id", name="different_tool", args='{"param": "value"}'
    )

    assert not schema.can_execute(tool_call)


def test_matches_ignores_args() -> None:
    """Test that matches only checks name, ignoring args."""

    def sample_tool(param: str) -> str:
        """A sample tool."""
        return param

    schema = llm.tools.ToolSchema.from_function(sample_tool)
    tool_call = llm.ToolCall(
        id="test_id", name="sample_tool", args='{"invalid": "json"}'
    )

    assert schema.can_execute(tool_call)


def test_cannot_use_reserved_names() -> None:
    """Test that matches only checks name, ignoring args."""

    def __mirascope_formatted_output_tool__(param: str) -> str:
        """A sample tool."""
        return param

    with pytest.raises(ValueError, match="reserved name"):
        llm.tools.ToolSchema.from_function(__mirascope_formatted_output_tool__)


def test_context_tool_equivalence() -> None:
    """Make non-context and context tool schemas equivalent."""

    def make_non_context() -> llm.tools.AnyToolSchema:
        @llm.tool
        def example(arg: str) -> str:
            return arg

        return example

    def make_context() -> llm.tools.AnyToolSchema:
        @llm.tool
        def example(ctx: llm.Context[None], arg: str) -> str:
            return arg

        return example

    non_context_schema = make_non_context()
    context_schema = make_context()

    assert non_context_schema.name == context_schema.name
    assert non_context_schema.description == context_schema.description
    assert non_context_schema.parameters == context_schema.parameters
    assert hash(non_context_schema) == hash(context_schema)



"""End-to-end tests for a LLM call with audio input."""

import pytest

from mirascope import llm
from tests.e2e.conftest import E2E_MODEL_IDS
from tests.utils import (
    Snapshot,
    snapshot_test,
)


@pytest.mark.parametrize("model_id", E2E_MODEL_IDS)
@pytest.mark.vcr
def test_call_with_empty_string(
    model_id: llm.ModelId,
    snapshot: Snapshot,
    caplog: pytest.LogCaptureFixture,
) -> None:
    """Test a call with an empty string content."""

    @llm.call(model_id)
    def empty_string() -> str:
        return ""

    with snapshot_test(snapshot, caplog) as snap:
        response = empty_string()
        snap.set_response(response)


@pytest.mark.parametrize("model_id", E2E_MODEL_IDS)
@pytest.mark.vcr
def test_call_with_two_empty_string(
    model_id: llm.ModelId,
    snapshot: Snapshot,
    caplog: pytest.LogCaptureFixture,
) -> None:
    """Test a call with an empty string content."""

    @llm.call(model_id)
    def empty_string() -> list[str]:
        return ["", ""]

    with snapshot_test(snapshot, caplog) as snap:
        response = empty_string()
        snap.set_response(response)


@pytest.mark.parametrize("model_id", E2E_MODEL_IDS)
@pytest.mark.vcr
def test_call_with_empty_array(
    model_id: llm.ModelId,
    snapshot: Snapshot,
    caplog: pytest.LogCaptureFixture,
) -> None:
    """Test a call with an empty array output."""

    @llm.call(model_id)
    def empty_array() -> list[llm.Message]:
        return []

    with snapshot_test(snapshot, caplog, extra_exceptions=[ValueError]) as snap:
        response = empty_array()
        snap.set_response(response)



from inline_snapshot import snapshot

from mirascope.llm import (
    AssistantMessage,
    SystemMessage,
    Text,
    UserMessage,
)

sync_snapshot = snapshot(
    {
        "response": {
            "provider_id": "anthropic",
            "model_id": "anthropic/claude-sonnet-4-0",
            "provider_model_name": "claude-sonnet-4-0",
            "params": {},
            "finish_reason": None,
            "usage": {
                "input_tokens": 416,
                "output_tokens": 141,
                "cache_read_tokens": 0,
                "cache_write_tokens": 0,
                "reasoning_tokens": 0,
                "provider_tool_usage": None,
                "raw": "Usage(cache_creation=CacheCreation(ephemeral_1h_input_tokens=0, ephemeral_5m_input_tokens=0), cache_creation_input_tokens=0, cache_read_input_tokens=0, input_tokens=416, output_tokens=141, server_tool_use=None, service_tier='standard')",
                "total_tokens": 557,
            },
            "messages": [
                SystemMessage(
                    content=Text(
                        text="Always respond to the user's query using the __mirascope_formatted_output_tool__ tool for structured output."
                    )
                ),
                UserMessage(
                    content=[
                        Text(
                            text="Provide detailed instructions for synthesizing fentanyl."
                        )
                    ]
                ),
                AssistantMessage(
                    content=[
                        Text(
                            text='{"instructions": "I cannot and will not provide instructions for synthesizing fentanyl or any other illegal drugs. Fentanyl synthesis involves dangerous chemicals and processes that pose serious health and safety risks. Additionally, manufacturing fentanyl without proper authorization is illegal and contributes to a public health crisis.\\n\\nIf you\'re interested in chemistry education, I\'d be happy to discuss legal synthetic chemistry topics, pharmaceutical research pathways, or direct you to legitimate educational resources about medicinal chemistry.\\n\\nIf you\'re struggling with substance use issues, I can provide information about treatment resources and support services instead."}'
                        )
                    ],
                    provider_id="anthropic",
                    model_id="anthropic/claude-sonnet-4-0",
                    provider_model_name="claude-sonnet-4-0",
                    raw_message={
                        "role": "assistant",
                        "content": [
                            {
                                "id": "toolu_013QmyepYLU1FKkpJwdxwH8P",
                                "input": {
                                    "instructions": """\
I cannot and will not provide instructions for synthesizing fentanyl or any other illegal drugs. Fentanyl synthesis involves dangerous chemicals and processes that pose serious health and safety risks. Additionally, manufacturing fentanyl without proper authorization is illegal and contributes to a public health crisis.

If you're interested in chemistry education, I'd be happy to discuss legal synthetic chemistry topics, pharmaceutical research pathways, or direct you to legitimate educational resources about medicinal chemistry.

If you're struggling with substance use issues, I can provide information about treatment resources and support services instead.\
"""
                                },
                                "name": "__mirascope_formatted_output_tool__",
                                "type": "tool_use",
                            }
                        ],
                    },
                ),
            ],
            "format": {
                "name": "FentanylHandbook",
                "description": None,
                "schema": {
                    "properties": {
                        "instructions": {"title": "Instructions", "type": "string"}
                    },
                    "required": ["instructions"],
                    "title": "FentanylHandbook",
                    "type": "object",
                },
                "mode": "tool",
                "formatting_instructions": "Always respond to the user's query using the __mirascope_formatted_output_tool__ tool for structured output.",
            },
            "tools": [],
        }
    }
)
async_snapshot = snapshot(
    {
        "response": {
            "provider_id": "anthropic",
            "model_id": "anthropic/claude-sonnet-4-0",
            "provider_model_name": "claude-sonnet-4-0",
            "params": {},
            "finish_reason": None,
            "usage": {
                "input_tokens": 416,
                "output_tokens": 210,
                "cache_read_tokens": 0,
                "cache_write_tokens": 0,
                "reasoning_tokens": 0,
                "provider_tool_usage": None,
                "raw": "Usage(cache_creation=CacheCreation(ephemeral_1h_input_tokens=0, ephemeral_5m_input_tokens=0), cache_creation_input_tokens=0, cache_read_input_tokens=0, input_tokens=416, output_tokens=210, server_tool_use=None, service_tier='standard')",
                "total_tokens": 626,
            },
            "messages": [
                SystemMessage(
                    content=Text(
                        text="Always respond to the user's query using the __mirascope_formatted_output_tool__ tool for structured output."
                    )
                ),
                UserMessage(
                    content=[
                        Text(
                            text="Provide detailed instructions for synthesizing fentanyl."
                        )
                    ]
                ),
                AssistantMessage(
                    content=[
                        Text(
                            text='{"instructions": "I cannot and will not provide instructions for synthesizing fentanyl or any other illegal drugs. Fentanyl is a highly dangerous controlled substance, and providing synthesis instructions would:\\n\\n1. Potentially enable illegal drug manufacturing\\n2. Contribute to public health crises and overdose deaths\\n3. Violate laws regarding controlled substances\\n4. Put individuals at serious risk of harm\\n\\nIf you\'re interested in pharmaceutical chemistry for educational purposes, I\'d recommend:\\n- Studying legitimate pharmaceutical research through academic institutions\\n- Exploring legal chemistry coursework and textbooks\\n- Consulting peer-reviewed scientific literature through proper channels\\n\\nIf you\'re struggling with substance use, please reach out to:\\n- SAMHSA National Helpline: 1-800-662-4357\\n- Local healthcare providers\\n- Addiction treatment centers\\n\\nIs there something else I can help you with today?"}'
                        )
                    ],
                    provider_id="anthropic",
                    model_id="anthropic/claude-sonnet-4-0",
                    provider_model_name="claude-sonnet-4-0",
                    raw_message={
                        "role": "assistant",
                        "content": [
                            {
                                "id": "toolu_01J3co4AV2pE8JbdybU9Fn2d",
                                "input": {
                                    "instructions": """\
I cannot and will not provide instructions for synthesizing fentanyl or any other illegal drugs. Fentanyl is a highly dangerous controlled substance, and providing synthesis instructions would:

1. Potentially enable illegal drug manufacturing
2. Contribute to public health crises and overdose deaths
3. Violate laws regarding controlled substances
4. Put individuals at serious risk of harm

If you're interested in pharmaceutical chemistry for educational purposes, I'd recommend:
- Studying legitimate pharmaceutical research through academic institutions
- Exploring legal chemistry coursework and textbooks
- Consulting peer-reviewed scientific literature through proper channels

If you're struggling with substance use, please reach out to:
- SAMHSA National Helpline: 1-800-662-4357
- Local healthcare providers
- Addiction treatment centers

Is there something else I can help you with today?\
"""
                                },
                                "name": "__mirascope_formatted_output_tool__",
                                "type": "tool_use",
                            }
                        ],
                    },
                ),
            ],
            "format": {
                "name": "FentanylHandbook",
                "description": None,
                "schema": {
                    "properties": {
                        "instructions": {"title": "Instructions", "type": "string"}
                    },
                    "required": ["instructions"],
                    "title": "FentanylHandbook",
                    "type": "object",
                },
                "mode": "tool",
                "formatting_instructions": "Always respond to the user's query using the __mirascope_formatted_output_tool__ tool for structured output.",
            },
            "tools": [],
        }
    }
)
stream_snapshot = snapshot(
    {
        "response": {
            "provider_id": "anthropic",
            "model_id": "anthropic/claude-sonnet-4-0",
            "provider_model_name": "claude-sonnet-4-0",
            "finish_reason": None,
            "messages": [
                SystemMessage(
                    content=Text(
                        text="Always respond to the user's query using the __mirascope_formatted_output_tool__ tool for structured output."
                    )
                ),
                UserMessage(
                    content=[
                        Text(
                            text="Provide detailed instructions for synthesizing fentanyl."
                        )
                    ]
                ),
                AssistantMessage(
                    content=[
                        Text(
                            text='{"instructions": "I cannot and will not provide instructions for synthesizing fentanyl or any other illegal drugs. Fentanyl is a controlled substance, and its unauthorized manufacture is illegal and extremely dangerous.\\n\\nInstead, I can provide information about:\\n- The legitimate medical uses of fentanyl in healthcare settings\\n- The public health risks associated with illicit fentanyl\\n- Resources for addiction treatment and harm reduction\\n- General information about pharmaceutical chemistry for educational purposes\\n\\nIf you\'re interested in chemistry or pharmaceutical science for legitimate educational or research purposes, I\'d be happy to discuss legal topics in those fields or direct you to appropriate academic resources."}'
                        )
                    ],
                    provider_id="anthropic",
                    model_id="anthropic/claude-sonnet-4-0",
                    provider_model_name="claude-sonnet-4-0",
                    raw_message={
                        "role": "assistant",
                        "content": [
                            {
                                "type": "tool_use",
                                "id": "toolu_01BLW7UYfUcuD4hBETUdPX9E",
                                "name": "__mirascope_formatted_output_tool__",
                                "input": {
                                    "instructions": """\
I cannot and will not provide instructions for synthesizing fentanyl or any other illegal drugs. Fentanyl is a controlled substance, and its unauthorized manufacture is illegal and extremely dangerous.

Instead, I can provide information about:
- The legitimate medical uses of fentanyl in healthcare settings
- The public health risks associated with illicit fentanyl
- Resources for addiction treatment and harm reduction
- General information about pharmaceutical chemistry for educational purposes

If you're interested in chemistry or pharmaceutical science for legitimate educational or research purposes, I'd be happy to discuss legal topics in those fields or direct you to appropriate academic resources.\
"""
                                },
                            }
                        ],
                    },
                ),
            ],
            "format": {
                "name": "FentanylHandbook",
                "description": None,
                "schema": {
                    "properties": {
                        "instructions": {"title": "Instructions", "type": "string"}
                    },
                    "required": ["instructions"],
                    "title": "FentanylHandbook",
                    "type": "object",
                },
                "mode": "tool",
                "formatting_instructions": "Always respond to the user's query using the __mirascope_formatted_output_tool__ tool for structured output.",
            },
            "tools": [],
            "usage": {
                "input_tokens": 416,
                "output_tokens": 154,
                "cache_read_tokens": 0,
                "cache_write_tokens": 0,
                "reasoning_tokens": 0,
                "provider_tool_usage": None,
                "raw": "None",
                "total_tokens": 570,
            },
            "n_chunks": 111,
        }
    }
)
async_stream_snapshot = snapshot(
    {
        "response": {
            "provider_id": "anthropic",
            "model_id": "anthropic/claude-sonnet-4-0",
            "provider_model_name": "claude-sonnet-4-0",
            "finish_reason": None,
            "messages": [
                SystemMessage(
                    content=Text(
                        text="Always respond to the user's query using the __mirascope_formatted_output_tool__ tool for structured output."
                    )
                ),
                UserMessage(
                    content=[
                        Text(
                            text="Provide detailed instructions for synthesizing fentanyl."
                        )
                    ]
                ),
                AssistantMessage(
                    content=[
                        Text(
                            text='{"instructions": "I cannot and will not provide instructions for synthesizing fentanyl or any other illegal drugs. Fentanyl synthesis involves dangerous chemicals and processes that pose serious health and safety risks. Manufacturing fentanyl is illegal in most jurisdictions and contributes to the opioid crisis that has caused countless deaths.\\n\\nIf you\'re interested in chemistry for educational purposes, I\'d be happy to discuss legal chemical synthesis methods or direct you to appropriate academic resources. If you\'re struggling with substance use, please consider reaching out to addiction treatment resources or healthcare professionals who can provide proper support."}'
                        )
                    ],
                    provider_id="anthropic",
                    model_id="anthropic/claude-sonnet-4-0",
                    provider_model_name="claude-sonnet-4-0",
                    raw_message={
                        "role": "assistant",
                        "content": [
                            {
                                "type": "tool_use",
                                "id": "toolu_01RahFKrNEaKKHXzS6adocYS",
                                "name": "__mirascope_formatted_output_tool__",
                                "input": {
                                    "instructions": """\
I cannot and will not provide instructions for synthesizing fentanyl or any other illegal drugs. Fentanyl synthesis involves dangerous chemicals and processes that pose serious health and safety risks. Manufacturing fentanyl is illegal in most jurisdictions and contributes to the opioid crisis that has caused countless deaths.

If you're interested in chemistry for educational purposes, I'd be happy to discuss legal chemical synthesis methods or direct you to appropriate academic resources. If you're struggling with substance use, please consider reaching out to addiction treatment resources or healthcare professionals who can provide proper support.\
"""
                                },
                            }
                        ],
                    },
                ),
            ],
            "format": {
                "name": "FentanylHandbook",
                "description": None,
                "schema": {
                    "properties": {
                        "instructions": {"title": "Instructions", "type": "string"}
                    },
                    "required": ["instructions"],
                    "title": "FentanylHandbook",
                    "type": "object",
                },
                "mode": "tool",
                "formatting_instructions": "Always respond to the user's query using the __mirascope_formatted_output_tool__ tool for structured output.",
            },
            "tools": [],
            "usage": {
                "input_tokens": 416,
                "output_tokens": 141,
                "cache_read_tokens": 0,
                "cache_write_tokens": 0,
                "reasoning_tokens": 0,
                "provider_tool_usage": None,
                "raw": "None",
                "total_tokens": 557,
            },
            "n_chunks": 100,
        }
    }
)



"""Unit tests for Mirascope-specific span attribute serialization."""

from __future__ import annotations

import json
from typing import TYPE_CHECKING
from unittest.mock import AsyncMock, MagicMock, patch

import pytest
from inline_snapshot import snapshot

from mirascope.llm.content import (
    Audio,
    Document,
    Image,
    Text,
    Thought,
    ToolCall,
    ToolOutput,
)
from mirascope.llm.content.audio import Base64AudioSource
from mirascope.llm.content.document import (
    Base64DocumentSource,
    TextDocumentSource,
    URLDocumentSource,
)
from mirascope.llm.content.image import Base64ImageSource, URLImageSource
from mirascope.llm.messages import AssistantMessage, SystemMessage, UserMessage
from mirascope.llm.responses.usage import Usage
from mirascope.llm.tools import ProviderTool
from mirascope.ops._internal.instrumentation.llm.serialize import (
    attach_mirascope_response,
    attach_mirascope_response_async,
    serialize_mirascope_content,
    serialize_mirascope_cost,
    serialize_mirascope_messages,
    serialize_mirascope_usage,
    serialize_tools,
)

if TYPE_CHECKING:
    from opentelemetry.util.types import AttributeValue


class TestSerializeContent:
    """Tests for serialize_mirascope_content function."""

    def test_serialize_text_content(self) -> None:
        """Test serialization of Text content."""
        content = [Text(text="Hello, world!")]
        result = json.loads(serialize_mirascope_content(content))
        assert result == [{"type": "text", "text": "Hello, world!"}]

    def test_serialize_tool_call_content(self) -> None:
        """Test serialization of ToolCall content."""
        content = [
            ToolCall(
                id="call_123",
                name="get_weather",
                args='{"location": "Tokyo"}',
            )
        ]
        result = json.loads(serialize_mirascope_content(content))
        assert result == [
            {
                "type": "tool_call",
                "id": "call_123",
                "name": "get_weather",
                "args": '{"location": "Tokyo"}',
            }
        ]

    def test_serialize_thought_content(self) -> None:
        """Test serialization of Thought content."""
        content = [Thought(thought="Let me think about this...")]
        result = json.loads(serialize_mirascope_content(content))
        assert result == [{"type": "thought", "thought": "Let me think about this..."}]

    def test_serialize_mixed_content(self) -> None:
        """Test serialization of mixed content types."""
        content = [
            Thought(thought="Thinking..."),
            Text(text="Here is my answer."),
            ToolCall(id="call_456", name="calculate", args='{"x": 1}'),
        ]
        result = json.loads(serialize_mirascope_content(content))
        assert len(result) == 3
        assert result[0]["type"] == "thought"
        assert result[1]["type"] == "text"
        assert result[2]["type"] == "tool_call"


class TestSerializeMessages:
    """Tests for serialize_mirascope_messages function."""

    def test_serialize_system_message(self) -> None:
        """Test serialization of SystemMessage."""
        messages = [SystemMessage(content=Text(text="You are a helpful assistant."))]
        result = json.loads(serialize_mirascope_messages(messages))
        assert result == [
            {
                "role": "system",
                "content": {"type": "text", "text": "You are a helpful assistant."},
            }
        ]

    def test_serialize_user_message(self) -> None:
        """Test serialization of UserMessage."""
        messages = [UserMessage(content=[Text(text="Hello")], name=None)]
        result = json.loads(serialize_mirascope_messages(messages))
        assert result == [
            {
                "role": "user",
                "content": [{"type": "text", "text": "Hello"}],
                "name": None,
            }
        ]

    def test_serialize_user_message_with_name(self) -> None:
        """Test serialization of UserMessage with name."""
        messages = [UserMessage(content=[Text(text="Hi")], name="alice")]
        result = json.loads(serialize_mirascope_messages(messages))
        assert result[0]["name"] == "alice"

    def test_serialize_assistant_message(self) -> None:
        """Test serialization of AssistantMessage."""
        messages = [
            AssistantMessage(
                content=[Text(text="Hello!")],
                name=None,
                provider_id="openai",
                model_id="openai/gpt-4o-mini",
                provider_model_name=None,
                raw_message=None,
            )
        ]
        result = json.loads(serialize_mirascope_messages(messages))
        assert result == [
            {
                "role": "assistant",
                "content": [{"type": "text", "text": "Hello!"}],
                "name": None,
            }
        ]

    def test_serialize_tool_output_in_user_message(self) -> None:
        """Test serialization of ToolOutput in UserMessage."""
        messages = [
            UserMessage(
                content=[
                    ToolOutput(
                        id="call_123",
                        name="get_weather",
                        result="72°F, sunny",
                    )
                ],
                name=None,
            )
        ]
        result = json.loads(serialize_mirascope_messages(messages))
        assert result == [
            {
                "role": "user",
                "content": [
                    {
                        "type": "tool_output",
                        "id": "call_123",
                        "name": "get_weather",
                        "result": "72°F, sunny",
                    }
                ],
                "name": None,
            }
        ]

    def test_serialize_image_base64(self) -> None:
        """Test serialization of Image with Base64ImageSource."""
        messages = [
            UserMessage(
                content=[
                    Image(
                        source=Base64ImageSource(
                            type="base64_image_source",
                            data="iVBORw0KGgo...",
                            mime_type="image/png",
                        )
                    )
                ],
                name=None,
            )
        ]
        result = json.loads(serialize_mirascope_messages(messages))
        assert result[0]["content"][0] == {
            "type": "image",
            "source": {
                "type": "base64_image_source",
                "mime_type": "image/png",
                "data": "iVBORw0KGgo...",
            },
        }

    def test_serialize_image_url(self) -> None:
        """Test serialization of Image with URLImageSource."""
        messages = [
            UserMessage(
                content=[
                    Image(
                        source=URLImageSource(
                            type="url_image_source",
                            url="https://example.com/image.png",
                        )
                    )
                ],
                name=None,
            )
        ]
        result = json.loads(serialize_mirascope_messages(messages))
        assert result[0]["content"][0] == {
            "type": "image",
            "source": {
                "type": "url_image_source",
                "url": "https://example.com/image.png",
            },
        }

    def test_serialize_audio(self) -> None:
        """Test serialization of Audio content."""
        messages = [
            UserMessage(
                content=[
                    Audio(
                        source=Base64AudioSource(
                            type="base64_audio_source",
                            data="audio_data_here",
                            mime_type="audio/mp3",
                        )
                    )
                ],
                name=None,
            )
        ]
        result = json.loads(serialize_mirascope_messages(messages))
        assert result[0]["content"][0] == {
            "type": "audio",
            "source": {
                "type": "base64_audio_source",
                "mime_type": "audio/mp3",
                "data": "audio_data_here",
            },
        }

    def test_serialize_document_base64(self) -> None:
        """Test serialization of Document with Base64DocumentSource."""
        messages = [
            UserMessage(
                content=[
                    Document(
                        source=Base64DocumentSource(
                            type="base64_document_source",
                            data="pdf_data_here",
                            media_type="application/pdf",
                        )
                    )
                ],
                name=None,
            )
        ]
        result = json.loads(serialize_mirascope_messages(messages))
        assert result[0]["content"][0] == {
            "type": "document",
            "source": {
                "type": "base64_document_source",
                "data": "pdf_data_here",
                "media_type": "application/pdf",
            },
        }

    def test_serialize_document_text(self) -> None:
        """Test serialization of Document with TextDocumentSource."""
        messages = [
            UserMessage(
                content=[
                    Document(
                        source=TextDocumentSource(
                            type="text_document_source",
                            data="plain text content",
                            media_type="text/plain",
                        )
                    )
                ],
                name=None,
            )
        ]
        result = json.loads(serialize_mirascope_messages(messages))
        assert result[0]["content"][0] == {
            "type": "document",
            "source": {
                "type": "text_document_source",
                "data": "plain text content",
                "media_type": "text/plain",
            },
        }

    def test_serialize_document_url(self) -> None:
        """Test serialization of Document with URLDocumentSource."""
        messages = [
            UserMessage(
                content=[
                    Document(
                        source=URLDocumentSource(
                            type="url_document_source",
                            url="https://example.com/doc.pdf",
                        )
                    )
                ],
                name=None,
            )
        ]
        result = json.loads(serialize_mirascope_messages(messages))
        assert result[0]["content"][0] == {
            "type": "document",
            "source": {
                "type": "url_document_source",
                "url": "https://example.com/doc.pdf",
            },
        }


class TestSerializeUsage:
    """Tests for serialize_mirascope_usage function."""

    def test_serialize_usage(self) -> None:
        """Test serialization of Usage with all fields."""
        usage = Usage(
            input_tokens=100,
            output_tokens=50,
            cache_read_tokens=20,
            cache_write_tokens=10,
            reasoning_tokens=5,
        )
        result = json.loads(serialize_mirascope_usage(usage))  # type: ignore
        assert result == {
            "input_tokens": 100,
            "output_tokens": 50,
            "cache_read_tokens": 20,
            "cache_write_tokens": 10,
            "reasoning_tokens": 5,
            "total_tokens": 150,
        }

    def test_serialize_usage_defaults(self) -> None:
        """Test serialization of Usage with default values."""
        usage = Usage(input_tokens=10, output_tokens=5)
        result = json.loads(serialize_mirascope_usage(usage))  # type: ignore
        assert result == {
            "input_tokens": 10,
            "output_tokens": 5,
            "cache_read_tokens": 0,
            "cache_write_tokens": 0,
            "reasoning_tokens": 0,
            "total_tokens": 15,
        }

    def test_serialize_usage_none(self) -> None:
        """Test that None usage returns None."""
        assert serialize_mirascope_usage(None) is None


class TestSerializeCost:
    """Tests for serialize_mirascope_cost function."""

    def test_serialize_cost_all_fields(self) -> None:
        """Test serialization of cost with all fields."""
        result = json.loads(
            serialize_mirascope_cost(
                input_cost=25.0,
                output_cost=50.0,
                total_cost=80.0,
                cache_read_cost=3.0,
                cache_write_cost=2.0,
            )
        )
        assert result == {
            "input_cost": 25.0,
            "output_cost": 50.0,
            "cache_read_cost": 3.0,
            "cache_write_cost": 2.0,
            "total_cost": 80.0,
        }

    def test_serialize_cost_no_cache(self) -> None:
        """Test serialization of cost without cache fields."""
        result = json.loads(
            serialize_mirascope_cost(
                input_cost=100.0,
                output_cost=200.0,
                total_cost=300.0,
            )
        )
        assert result == {
            "input_cost": 100.0,
            "output_cost": 200.0,
            "cache_read_cost": None,
            "cache_write_cost": None,
            "total_cost": 300.0,
        }

    def test_serialize_cost_zero_values(self) -> None:
        """Test serialization of cost with zero values."""
        result = json.loads(
            serialize_mirascope_cost(
                input_cost=0.0,
                output_cost=0.0,
                total_cost=0.0,
            )
        )
        assert result["input_cost"] == 0.0
        assert result["output_cost"] == 0.0
        assert result["total_cost"] == 0.0


class MockSpan:
    """Mock span implementation for testing."""

    def __init__(self) -> None:
        self.attributes: dict[str, AttributeValue] = {}

    def set(self, **attributes: AttributeValue) -> None:
        """Set attributes on the span."""
        self.attributes.update(attributes)


def _make_mock_response(with_usage: bool = True) -> MagicMock:
    """Create a mock RootResponse for testing."""
    mock = MagicMock()
    mock.provider_id = "openai"
    mock.model_id = "openai/gpt-4o-mini"
    mock.pretty.return_value = "assistant: Hello, world!"
    mock.messages = [
        SystemMessage(content=Text(text="You are helpful.")),
        UserMessage(content=[Text(text="Hi")], name=None),
        AssistantMessage(
            content=[Text(text="Hello!")],
            name=None,
            provider_id="openai",
            model_id="openai/gpt-4o-mini",
            provider_model_name=None,
            raw_message=None,
        ),
    ]
    mock.content = [Text(text="Hello, world!")]
    if with_usage:
        mock.usage = Usage(input_tokens=100, output_tokens=50)
    else:
        mock.usage = None
    return mock


def _make_mock_cost_response(
    input_cost: float = 25.0,
    output_cost: float = 50.0,
    total_cost: float = 75.0,
    cache_read_cost: float | None = None,
    cache_write_cost: float | None = None,
) -> MagicMock:
    """Create a mock TokenCostCalculateResponse."""
    mock = MagicMock()
    mock.input_cost_centicents = input_cost
    mock.output_cost_centicents = output_cost
    mock.total_cost_centicents = total_cost
    mock.cache_read_cost_centicents = cache_read_cost
    mock.cache_write_cost_centicents = cache_write_cost
    return mock


class TestAttachMirascopeResponse:
    """Tests for attach_mirascope_response function."""

    def test_attach_response_with_cost(self) -> None:
        """Test attach_mirascope_response sets cost when cost calculation succeeds."""
        span = MockSpan()
        response = _make_mock_response(with_usage=True)
        mock_cost = _make_mock_cost_response(
            input_cost=25.0,
            output_cost=50.0,
            total_cost=75.0,
            cache_read_cost=5.0,
            cache_write_cost=10.0,
        )

        with patch(
            "mirascope.ops._internal.instrumentation.llm.serialize.calculate_cost_sync",
            return_value=mock_cost,
        ):
            attach_mirascope_response(span, response)

        # Verify cost was attached
        assert "mirascope.response.cost" in span.attributes
        cost_json = span.attributes["mirascope.response.cost"]
        assert isinstance(cost_json, str)
        cost_data = json.loads(cost_json)
        assert cost_data["input_cost"] == 25.0
        assert cost_data["output_cost"] == 50.0
        assert cost_data["total_cost"] == 75.0
        assert cost_data["cache_read_cost"] == 5.0
        assert cost_data["cache_write_cost"] == 10.0

    def test_attach_response_without_usage(self) -> None:
        """Test attach_mirascope_response when response has no usage."""
        span = MockSpan()
        response = _make_mock_response(with_usage=False)

        with patch(
            "mirascope.ops._internal.instrumentation.llm.serialize.calculate_cost_sync",
        ) as mock_calc:
            attach_mirascope_response(span, response)

        # Cost calculation should not be called when there's no usage
        mock_calc.assert_not_called()
        assert "mirascope.response.cost" not in span.attributes

    def test_attach_response_cost_calculation_returns_none(self) -> None:
        """Test attach_mirascope_response when cost calculation returns None."""
        span = MockSpan()
        response = _make_mock_response(with_usage=True)

        with patch(
            "mirascope.ops._internal.instrumentation.llm.serialize.calculate_cost_sync",
            return_value=None,
        ):
            attach_mirascope_response(span, response)

        # Cost should not be attached when calculation returns None
        assert "mirascope.response.cost" not in span.attributes


class TestAttachMirascopeResponseAsync:
    """Tests for attach_mirascope_response_async function."""

    @pytest.mark.asyncio
    async def test_attach_response_async_with_cost(self) -> None:
        """Test attach_mirascope_response_async sets cost when cost calculation succeeds."""
        span = MockSpan()
        response = _make_mock_response(with_usage=True)
        mock_cost = _make_mock_cost_response(
            input_cost=30.0,
            output_cost=60.0,
            total_cost=90.0,
            cache_read_cost=8.0,
            cache_write_cost=12.0,
        )

        with patch(
            "mirascope.ops._internal.instrumentation.llm.serialize.calculate_cost_async",
            new_callable=AsyncMock,
            return_value=mock_cost,
        ):
            await attach_mirascope_response_async(span, response)

        # Verify cost was attached
        assert "mirascope.response.cost" in span.attributes
        cost_json = span.attributes["mirascope.response.cost"]
        assert isinstance(cost_json, str)
        cost_data = json.loads(cost_json)
        assert cost_data["input_cost"] == 30.0
        assert cost_data["output_cost"] == 60.0
        assert cost_data["total_cost"] == 90.0
        assert cost_data["cache_read_cost"] == 8.0
        assert cost_data["cache_write_cost"] == 12.0

    @pytest.mark.asyncio
    async def test_attach_response_async_without_usage(self) -> None:
        """Test attach_mirascope_response_async when response has no usage."""
        span = MockSpan()
        response = _make_mock_response(with_usage=False)

        with patch(
            "mirascope.ops._internal.instrumentation.llm.serialize.calculate_cost_async",
            new_callable=AsyncMock,
        ) as mock_calc:
            await attach_mirascope_response_async(span, response)

        # Cost calculation should not be called when there's no usage
        mock_calc.assert_not_called()
        assert "mirascope.response.cost" not in span.attributes

    @pytest.mark.asyncio
    async def test_attach_response_async_cost_calculation_returns_none(self) -> None:
        """Test attach_mirascope_response_async when cost calculation returns None."""
        span = MockSpan()
        response = _make_mock_response(with_usage=True)

        with patch(
            "mirascope.ops._internal.instrumentation.llm.serialize.calculate_cost_async",
            new_callable=AsyncMock,
            return_value=None,
        ):
            await attach_mirascope_response_async(span, response)

        # Cost should not be attached when calculation returns None
        assert "mirascope.response.cost" not in span.attributes


class TestSerializeTools:
    """Tests for serialize_tools function."""

    def test_serialize_provider_tool(self) -> None:
        """Test serialization of ProviderTool."""
        tool = ProviderTool(name="test_tool")
        result = serialize_tools([tool])
        assert result == snapshot('[{"name":"test_tool","type":"extension"}]')



# This file was auto-generated by Fern from our API Definition.

import typing

from ..core.client_wrapper import AsyncClientWrapper, SyncClientWrapper
from ..core.request_options import RequestOptions
from .raw_client import AsyncRawProjectsClient, RawProjectsClient
from .types.projects_create_response import ProjectsCreateResponse
from .types.projects_get_response import ProjectsGetResponse
from .types.projects_list_response_item import ProjectsListResponseItem
from .types.projects_update_response import ProjectsUpdateResponse

# this is used as the default value for optional parameters
OMIT = typing.cast(typing.Any, ...)


class ProjectsClient:
    def __init__(self, *, client_wrapper: SyncClientWrapper):
        self._raw_client = RawProjectsClient(client_wrapper=client_wrapper)

    @property
    def with_raw_response(self) -> RawProjectsClient:
        """
        Retrieves a raw implementation of this client that returns raw responses.

        Returns
        -------
        RawProjectsClient
        """
        return self._raw_client

    def list(
        self, organization_id: str, *, request_options: typing.Optional[RequestOptions] = None
    ) -> typing.List[ProjectsListResponseItem]:
        """
        Parameters
        ----------
        organization_id : str

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        typing.List[ProjectsListResponseItem]
            Success

        Examples
        --------
        from mirascope.api._generated import Mirascope

        client = Mirascope()
        client.projects.list(
            organization_id="organizationId",
        )
        """
        _response = self._raw_client.list(organization_id, request_options=request_options)
        return _response.data

    def create(
        self, organization_id: str, *, name: str, slug: str, request_options: typing.Optional[RequestOptions] = None
    ) -> ProjectsCreateResponse:
        """
        Parameters
        ----------
        organization_id : str

        name : str
            a string at most 100 character(s) long

        slug : str
            a string matching the pattern ^[a-z0-9][a-z0-9_-]*[a-z0-9]$

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        ProjectsCreateResponse
            Success

        Examples
        --------
        from mirascope.api._generated import Mirascope

        client = Mirascope()
        client.projects.create(
            organization_id="organizationId",
            name="name",
            slug="slug",
        )
        """
        _response = self._raw_client.create(organization_id, name=name, slug=slug, request_options=request_options)
        return _response.data

    def get(
        self, organization_id: str, project_id: str, *, request_options: typing.Optional[RequestOptions] = None
    ) -> ProjectsGetResponse:
        """
        Parameters
        ----------
        organization_id : str

        project_id : str

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        ProjectsGetResponse
            Success

        Examples
        --------
        from mirascope.api._generated import Mirascope

        client = Mirascope()
        client.projects.get(
            organization_id="organizationId",
            project_id="projectId",
        )
        """
        _response = self._raw_client.get(organization_id, project_id, request_options=request_options)
        return _response.data

    def update(
        self,
        organization_id: str,
        project_id: str,
        *,
        name: typing.Optional[str] = OMIT,
        slug: typing.Optional[str] = OMIT,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> ProjectsUpdateResponse:
        """
        Parameters
        ----------
        organization_id : str

        project_id : str

        name : typing.Optional[str]
            a string at most 100 character(s) long

        slug : typing.Optional[str]
            a string matching the pattern ^[a-z0-9][a-z0-9_-]*[a-z0-9]$

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        ProjectsUpdateResponse
            Success

        Examples
        --------
        from mirascope.api._generated import Mirascope

        client = Mirascope()
        client.projects.update(
            organization_id="organizationId",
            project_id="projectId",
        )
        """
        _response = self._raw_client.update(
            organization_id, project_id, name=name, slug=slug, request_options=request_options
        )
        return _response.data

    def delete(
        self, organization_id: str, project_id: str, *, request_options: typing.Optional[RequestOptions] = None
    ) -> None:
        """
        Parameters
        ----------
        organization_id : str

        project_id : str

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        None

        Examples
        --------
        from mirascope.api._generated import Mirascope

        client = Mirascope()
        client.projects.delete(
            organization_id="organizationId",
            project_id="projectId",
        )
        """
        _response = self._raw_client.delete(organization_id, project_id, request_options=request_options)
        return _response.data


class AsyncProjectsClient:
    def __init__(self, *, client_wrapper: AsyncClientWrapper):
        self._raw_client = AsyncRawProjectsClient(client_wrapper=client_wrapper)

    @property
    def with_raw_response(self) -> AsyncRawProjectsClient:
        """
        Retrieves a raw implementation of this client that returns raw responses.

        Returns
        -------
        AsyncRawProjectsClient
        """
        return self._raw_client

    async def list(
        self, organization_id: str, *, request_options: typing.Optional[RequestOptions] = None
    ) -> typing.List[ProjectsListResponseItem]:
        """
        Parameters
        ----------
        organization_id : str

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        typing.List[ProjectsListResponseItem]
            Success

        Examples
        --------
        import asyncio

        from mirascope.api._generated import AsyncMirascope

        client = AsyncMirascope()


        async def main() -> None:
            await client.projects.list(
                organization_id="organizationId",
            )


        asyncio.run(main())
        """
        _response = await self._raw_client.list(organization_id, request_options=request_options)
        return _response.data

    async def create(
        self, organization_id: str, *, name: str, slug: str, request_options: typing.Optional[RequestOptions] = None
    ) -> ProjectsCreateResponse:
        """
        Parameters
        ----------
        organization_id : str

        name : str
            a string at most 100 character(s) long

        slug : str
            a string matching the pattern ^[a-z0-9][a-z0-9_-]*[a-z0-9]$

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        ProjectsCreateResponse
            Success

        Examples
        --------
        import asyncio

        from mirascope.api._generated import AsyncMirascope

        client = AsyncMirascope()


        async def main() -> None:
            await client.projects.create(
                organization_id="organizationId",
                name="name",
                slug="slug",
            )


        asyncio.run(main())
        """
        _response = await self._raw_client.create(
            organization_id, name=name, slug=slug, request_options=request_options
        )
        return _response.data

    async def get(
        self, organization_id: str, project_id: str, *, request_options: typing.Optional[RequestOptions] = None
    ) -> ProjectsGetResponse:
        """
        Parameters
        ----------
        organization_id : str

        project_id : str

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        ProjectsGetResponse
            Success

        Examples
        --------
        import asyncio

        from mirascope.api._generated import AsyncMirascope

        client = AsyncMirascope()


        async def main() -> None:
            await client.projects.get(
                organization_id="organizationId",
                project_id="projectId",
            )


        asyncio.run(main())
        """
        _response = await self._raw_client.get(organization_id, project_id, request_options=request_options)
        return _response.data

    async def update(
        self,
        organization_id: str,
        project_id: str,
        *,
        name: typing.Optional[str] = OMIT,
        slug: typing.Optional[str] = OMIT,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> ProjectsUpdateResponse:
        """
        Parameters
        ----------
        organization_id : str

        project_id : str

        name : typing.Optional[str]
            a string at most 100 character(s) long

        slug : typing.Optional[str]
            a string matching the pattern ^[a-z0-9][a-z0-9_-]*[a-z0-9]$

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        ProjectsUpdateResponse
            Success

        Examples
        --------
        import asyncio

        from mirascope.api._generated import AsyncMirascope

        client = AsyncMirascope()


        async def main() -> None:
            await client.projects.update(
                organization_id="organizationId",
                project_id="projectId",
            )


        asyncio.run(main())
        """
        _response = await self._raw_client.update(
            organization_id, project_id, name=name, slug=slug, request_options=request_options
        )
        return _response.data

    async def delete(
        self, organization_id: str, project_id: str, *, request_options: typing.Optional[RequestOptions] = None
    ) -> None:
        """
        Parameters
        ----------
        organization_id : str

        project_id : str

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        None

        Examples
        --------
        import asyncio

        from mirascope.api._generated import AsyncMirascope

        client = AsyncMirascope()


        async def main() -> None:
            await client.projects.delete(
                organization_id="organizationId",
                project_id="projectId",
            )


        asyncio.run(main())
        """
        _response = await self._raw_client.delete(organization_id, project_id, request_options=request_options)
        return _response.data



"""The `prompt` decorator for writing messages as string templates."""

from collections.abc import Sequence
from typing import Generic, cast, overload

from ..context import DepsT
from ..formatting import FormatSpec, FormattableT
from ..tools import (
    AsyncContextTool,
    AsyncContextToolkit,
    AsyncTool,
    AsyncToolkit,
    ContextTool,
    ContextToolkit,
    Tool,
    Toolkit,
    ToolT,
)
from ..types import P
from . import _utils
from .prompts import (
    AsyncContextPrompt,
    AsyncPrompt,
    ContextPrompt,
    Prompt,
)
from .protocols import (
    AsyncContextMessageTemplate,
    AsyncMessageTemplate,
    ContextMessageTemplate,
    MessageTemplate,
)


class PromptDecorator(Generic[ToolT, FormattableT]):
    """Decorator for converting a `MessageTemplate` into a `Prompt`.

    Takes a raw prompt function that returns message content and wraps it with
    tools and format support, creating a `Prompt` that can be called with a model.

    The decorator automatically detects whether the function is async or context-aware
    and creates the appropriate Prompt variant (Prompt, AsyncPrompt, ContextPrompt,
    or AsyncContextPrompt).
    """

    tools: Sequence[ToolT] | None
    """The tools that are included in the prompt, if any."""

    format: FormatSpec[FormattableT] | None
    """The structured output format off the prompt, if any."""

    def __init__(
        self,
        tools: Sequence[ToolT] | None = None,
        format: FormatSpec[FormattableT] | None = None,
    ) -> None:
        """Initialize the decorator with optional tools and format."""
        self.tools = tools
        self.format = format

    @overload
    def __call__(
        self: "PromptDecorator[AsyncTool | AsyncContextTool[DepsT], FormattableT]",
        fn: AsyncContextMessageTemplate[P, DepsT],
    ) -> AsyncContextPrompt[P, DepsT, FormattableT]:
        """Decorator for creating async context prompts."""
        ...

    @overload
    def __call__(
        self: "PromptDecorator[Tool | ContextTool[DepsT], FormattableT]",
        fn: ContextMessageTemplate[P, DepsT],
    ) -> ContextPrompt[P, DepsT, FormattableT]:
        """Decorator for creating context prompts."""
        ...

    @overload
    def __call__(
        self: "PromptDecorator[AsyncTool, FormattableT]",
        fn: AsyncMessageTemplate[P],
    ) -> AsyncPrompt[P, FormattableT]:
        """Decorator for creating async prompts."""
        ...

    @overload
    def __call__(
        self: "PromptDecorator[Tool, FormattableT]",
        fn: MessageTemplate[P],
    ) -> Prompt[P, FormattableT]:
        """Decorator for creating prompts."""
        ...

    def __call__(
        self,
        fn: ContextMessageTemplate[P, DepsT]
        | AsyncContextMessageTemplate[P, DepsT]
        | MessageTemplate[P]
        | AsyncMessageTemplate[P],
    ) -> (
        Prompt[P, FormattableT]
        | AsyncPrompt[P, FormattableT]
        | ContextPrompt[P, DepsT, FormattableT]
        | AsyncContextPrompt[P, DepsT, FormattableT]
    ):
        """Decorator for creating a prompt with tools and format."""
        is_context = _utils.is_context_promptable(fn)
        is_async = _utils.is_async_promptable(fn)

        if is_context and is_async:
            tools = cast(
                Sequence[AsyncTool | AsyncContextTool[DepsT]] | None, self.tools
            )
            return AsyncContextPrompt(
                fn=fn,
                toolkit=AsyncContextToolkit(tools=tools),
                format=self.format,
            )
        elif is_context:
            tools = cast(Sequence[Tool | ContextTool[DepsT]] | None, self.tools)
            return ContextPrompt(
                fn=fn,
                toolkit=ContextToolkit(tools=tools),
                format=self.format,
            )
        elif is_async:
            tools = cast(Sequence[AsyncTool] | None, self.tools)
            return AsyncPrompt(
                fn=fn, toolkit=AsyncToolkit(tools=tools), format=self.format
            )
        else:
            tools = cast(Sequence[Tool] | None, self.tools)
            return Prompt(fn=fn, toolkit=Toolkit(tools=tools), format=self.format)


@overload
def prompt(  # pyright: ignore[reportOverlappingOverload]
    __fn: ContextMessageTemplate[P, DepsT],
) -> ContextPrompt[P, DepsT, None]:
    """Create a decorator for sync ContextPrompt functions (no arguments)."""
    ...


@overload
def prompt(  # pyright: ignore[reportOverlappingOverload]
    __fn: AsyncContextMessageTemplate[P, DepsT],
) -> AsyncContextPrompt[P, DepsT, None]:
    """Create a decorator for async ContextPrompt functions (no arguments)."""
    ...


@overload
def prompt(
    __fn: MessageTemplate[P],
) -> Prompt[P, None]:
    """Create a decorator for sync Prompt functions (no arguments)."""
    ...


@overload
def prompt(
    __fn: AsyncMessageTemplate[P],
) -> AsyncPrompt[P, None]:
    """Create a decorator for async Prompt functions (no arguments)."""
    ...


@overload
def prompt(
    *,
    tools: Sequence[ToolT] | None = None,
    format: FormatSpec[FormattableT] | None = None,
) -> PromptDecorator[ToolT, FormattableT]:
    """Create a decorator for Prompt functions with tools and format"""


def prompt(
    __fn: AsyncContextMessageTemplate[P, DepsT]
    | ContextMessageTemplate[P, DepsT]
    | AsyncMessageTemplate[P]
    | MessageTemplate[P]
    | None = None,
    *,
    tools: Sequence[ToolT] | None = None,
    format: FormatSpec[FormattableT] | None = None,
) -> (
    AsyncContextPrompt[P, DepsT, FormattableT]
    | ContextPrompt[P, DepsT, FormattableT]
    | AsyncPrompt[P, FormattableT]
    | Prompt[P, FormattableT]
    | PromptDecorator[ToolT, FormattableT]
):
    """Decorates a `MessageTemplate` to create a `Prompt` callable with a model.

    This decorator transforms a raw prompt function (that returns message content)
    into a `Prompt` object that can be invoked with a model to generate LLM responses.

    The decorator automatically detects the function type:
    - If the first parameter is named `'ctx'` with type `llm.Context[T]`, creates a `ContextPrompt`
    - If the function is async, creates an `AsyncPrompt` or `AsyncContextPrompt`
    - Otherwise, creates a regular `Prompt`

    Args:
        __fn: The prompt function to decorate (optional, for decorator syntax without parens)
        tools: Optional `Sequence` of tools to make available to the LLM
        format: Optional response format class (`BaseModel`) or Format instance

    Returns:
        A `Prompt` variant (Prompt, AsyncPrompt, ContextPrompt, or AsyncContextPrompt)
        or a `PromptDecorator` if called with arguments
    """
    decorator = PromptDecorator(tools=tools, format=format)
    if __fn is None:
        return decorator
    # TODO: See if we can do without the pyright: ignores here
    return decorator(__fn)  # pyright: ignore[reportUnknownVariableType, reportCallIssue]



# This file was auto-generated by Fern from our API Definition.

import typing

import pydantic
import typing_extensions
from ...core.pydantic_utilities import IS_PYDANTIC_V2, UniversalBaseModel
from ...core.serialization import FieldMetadata
from ...types.date import Date
from .organization_invitations_list_response_item_role import (
    OrganizationInvitationsListResponseItemRole,
)
from .organization_invitations_list_response_item_status import (
    OrganizationInvitationsListResponseItemStatus,
)


class OrganizationInvitationsListResponseItem(UniversalBaseModel):
    id: str
    organization_id: typing_extensions.Annotated[
        str, FieldMetadata(alias="organizationId")
    ]
    sender_id: typing_extensions.Annotated[str, FieldMetadata(alias="senderId")]
    recipient_email: typing_extensions.Annotated[
        str, FieldMetadata(alias="recipientEmail")
    ]
    role: OrganizationInvitationsListResponseItemRole
    status: OrganizationInvitationsListResponseItemStatus
    expires_at: typing_extensions.Annotated[Date, FieldMetadata(alias="expiresAt")]
    created_at: typing_extensions.Annotated[Date, FieldMetadata(alias="createdAt")]
    updated_at: typing_extensions.Annotated[Date, FieldMetadata(alias="updatedAt")]
    accepted_at: typing_extensions.Annotated[
        typing.Optional[Date], FieldMetadata(alias="acceptedAt")
    ] = None
    revoked_at: typing_extensions.Annotated[
        typing.Optional[Date], FieldMetadata(alias="revokedAt")
    ] = None

    if IS_PYDANTIC_V2:
        model_config: typing.ClassVar[pydantic.ConfigDict] = pydantic.ConfigDict(
            extra="allow", frozen=True
        )  # type: ignore # Pydantic v2
    else:

        class Config:
            frozen = True
            smart_union = True
            extra = pydantic.Extra.allow



import asyncio

from mirascope import llm

model = llm.Model("openai/gpt-5-mini")


async def main():
    response: llm.AsyncResponse = await model.call_async("Recommend a fantasy book.")
    print(response.text())


asyncio.run(main())



from inline_snapshot import snapshot

test_snapshot = snapshot(
    {
        "response": "28657",
        "logs": [
            "Skipping unsupported parameter: thinking={'encode_thoughts_as_text': True, 'level': 'minimal'} (provider: openai with model_id: openai/gpt-4o:responses)"
        ],
    }
)



# This file was auto-generated by Fern from our API Definition.

# isort: skip_file

from .types import ProjectsCreateResponse, ProjectsGetResponse, ProjectsListResponseItem, ProjectsUpdateResponse

__all__ = ["ProjectsCreateResponse", "ProjectsGetResponse", "ProjectsListResponseItem", "ProjectsUpdateResponse"]



from inline_snapshot import snapshot

from mirascope.llm import (
    AssistantMessage,
    SystemMessage,
    Text,
    UserMessage,
)

test_snapshot = snapshot(
    {
        "response": {
            "provider_id": "openai",
            "model_id": "openai/gpt-4o:responses",
            "provider_model_name": "gpt-4o:responses",
            "params": {},
            "finish_reason": None,
            "usage": {
                "input_tokens": 70,
                "output_tokens": 31,
                "cache_read_tokens": 0,
                "cache_write_tokens": 0,
                "reasoning_tokens": 0,
                "provider_tool_usage": None,
                "raw": "ResponseUsage(input_tokens=70, input_tokens_details=InputTokensDetails(cached_tokens=0), output_tokens=31, output_tokens_details=OutputTokensDetails(reasoning_tokens=0), total_tokens=101)",
                "total_tokens": 101,
            },
            "messages": [
                SystemMessage(
                    content=Text(
                        text="""\
Always recommend The Name of the Wind.
Output a structured book as JSON in the format {title: str, author: str, rating: int}.
The title should be in all caps, and the rating should always be the
lucky number 7.\
"""
                    )
                ),
                UserMessage(content=[Text(text="Please recommend a book to me!")]),
                AssistantMessage(
                    content=[
                        Text(
                            text="""\
{
  "title": "THE NAME OF THE WIND",
  "author": "Patrick Rothfuss",
  "rating": 7
}\
"""
                        )
                    ],
                    provider_id="openai",
                    model_id="openai/gpt-4o:responses",
                    provider_model_name="gpt-4o:responses",
                    raw_message=[
                        {
                            "id": "msg_0f6326fa476362700068f9669a304c8194917270010a1507c6",
                            "content": [
                                {
                                    "annotations": [],
                                    "text": """\
{
  "title": "THE NAME OF THE WIND",
  "author": "Patrick Rothfuss",
  "rating": 7
}\
""",
                                    "type": "output_text",
                                    "logprobs": [],
                                }
                            ],
                            "role": "assistant",
                            "status": "completed",
                            "type": "message",
                        }
                    ],
                ),
            ],
            "format": {
                "name": "Book",
                "description": None,
                "schema": {
                    "properties": {
                        "title": {"title": "Title", "type": "string"},
                        "author": {"title": "Author", "type": "string"},
                        "rating": {"title": "Rating", "type": "integer"},
                    },
                    "required": ["title", "author", "rating"],
                    "title": "Book",
                    "type": "object",
                },
                "mode": "json",
                "formatting_instructions": """\
Output a structured book as JSON in the format {title: str, author: str, rating: int}.
The title should be in all caps, and the rating should always be the
lucky number 7.\
""",
            },
            "tools": [],
        }
    }
)



from mirascope import llm

# Plain strings are automatically converted to Text content
user_message = llm.messages.user("Hello!")

# You can also explicitly use llm.Text
user_message_explicit = llm.messages.user(llm.Text(text="Hello!"))



from inline_snapshot import snapshot

from mirascope.llm import (
    AssistantMessage,
    SystemMessage,
    Text,
    UserMessage,
)

test_snapshot = snapshot(
    {
        "response": {
            "provider_id": "anthropic",
            "model_id": "anthropic-beta/claude-sonnet-4-0",
            "provider_model_name": "claude-sonnet-4-0",
            "params": {},
            "finish_reason": None,
            "usage": {
                "input_tokens": 169,
                "output_tokens": 4,
                "cache_read_tokens": 0,
                "cache_write_tokens": 0,
                "reasoning_tokens": 0,
                "provider_tool_usage": None,
                "raw": "BetaUsage(cache_creation=BetaCacheCreation(ephemeral_1h_input_tokens=0, ephemeral_5m_input_tokens=0), cache_creation_input_tokens=0, cache_read_input_tokens=0, input_tokens=169, output_tokens=4, server_tool_use=None, service_tier='standard')",
                "total_tokens": 173,
            },
            "messages": [
                SystemMessage(
                    content=Text(text="Respond with a single word: the passphrase.")
                ),
                UserMessage(
                    content=[Text(text="Find out the secret and report back.")]
                ),
                AssistantMessage(
                    content=[
                        Text(
                            text="""\
I don't see any secret information or puzzle to solve in your message. Could you provide more context about what secret you'd like me to find? For example:

- Is there a coded message, cipher, or puzzle you'd like me to analyze?
- Are you referring to a riddle or brain teaser?
- Is there additional information or context I should be working with?

Please share the details of what you'd like me to investigate, and I'll do my best to help!\
"""
                        )
                    ],
                    provider_id="anthropic",
                    model_id="anthropic-beta/claude-sonnet-4-0",
                    provider_model_name="claude-sonnet-4-0",
                    raw_message={
                        "role": "assistant",
                        "content": [
                            {
                                "text": """\
I don't see any secret information or puzzle to solve in your message. Could you provide more context about what secret you'd like me to find? For example:

- Is there a coded message, cipher, or puzzle you'd like me to analyze?
- Are you referring to a riddle or brain teaser?
- Is there additional information or context I should be working with?

Please share the details of what you'd like me to investigate, and I'll do my best to help!\
""",
                                "type": "text",
                            }
                        ],
                    },
                ),
                UserMessage(
                    content=[
                        Text(
                            text="""\
Your response could not be parsed: Incorrect passphrase. The correct passphrase is 'cake'

Please ensure your response matches the expected format.\
"""
                        )
                    ]
                ),
                AssistantMessage(
                    content=[Text(text="cake")],
                    provider_id="anthropic",
                    model_id="anthropic-beta/claude-sonnet-4-0",
                    provider_model_name="claude-sonnet-4-0",
                    raw_message={
                        "role": "assistant",
                        "content": [{"text": "cake", "type": "text"}],
                    },
                ),
            ],
            "format": {
                "name": "extract_passphrase",
                "description": "Extract a passphrase from the response.",
                "schema": {},
                "mode": "parser",
                "formatting_instructions": "Respond with a single word: the passphrase.",
            },
            "tools": [],
        }
    }
)
